# Copyright 2012 OpenStack Foundation.
# Copyright 2013 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from unittest import mock
import uuid

from oslo_config import cfg
from oslo_utils.fixture import uuidsentinel as uuids
from oslo_utils import timeutils
from sqlalchemy import orm as sa_orm

from glance.common import crypt
from glance.common import exception
import glance.context
import glance.db
from glance.db.sqlalchemy import api
import glance.tests.unit.utils as unit_test_utils
import glance.tests.utils as test_utils

CONF = cfg.CONF
CONF.import_opt('metadata_encryption_key', 'glance.common.config')


UUID1 = 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d'
UUID2 = 'a85abd86-55b3-4d5b-b0b4-5d0a6e6042fc'
UUID3 = '971ec09a-8067-4bc8-a91f-ae3557f1c4c7'
UUID4 = '6bbe7cc2-eae7-4c0f-b50d-a7160b0c6a86'

TENANT1 = '6838eb7b-6ded-434a-882c-b344c77fe8df'
TENANT2 = '2c014f32-55eb-467d-8fcb-4bd706012f81'
TENANT3 = '5a3e60e8-cfa9-4a9e-a90a-62b42cea92b8'
TENANT4 = 'c6c87f25-8a94-47ed-8c83-053c25f42df4'

USER1 = '54492ba0-f4df-4e4e-be62-27f4d76b29cf'

UUID1_LOCATION = 'file:///path/to/image'
UUID1_LOCATION_METADATA = {'key': 'value'}
UUID3_LOCATION = 'http://somehost.com/place'

CHECKSUM = '93264c3edf5972c9f1cb309543d38a5c'
CHCKSUM1 = '43264c3edf4972c9f1cb309543d38a55'


TASK_ID_1 = 'b3006bd0-461e-4228-88ea-431c14e918b4'
TASK_ID_2 = '07b6b562-6770-4c8b-a649-37a515144ce9'
TASK_ID_3 = '72d16bb6-4d70-48a5-83fe-14bb842dc737'

NODE_REFERENCE_ID_1 = 1
NODE_REFERENCE_ID_2 = 2


def _db_fixture(id, **kwargs):
    obj = {
        'id': id,
        'name': None,
        'is_public': False,
        'properties': {},
        'checksum': None,
        'owner': None,
        'status': 'queued',
        'tags': [],
        'size': None,
        'locations': [],
        'protected': False,
        'disk_format': None,
        'container_format': None,
        'deleted': False,
        'min_ram': None,
        'min_disk': None,
    }
    if 'visibility' in kwargs:
        obj.pop('is_public')
    obj.update(kwargs)
    return obj


def _db_image_member_fixture(image_id, member_id, **kwargs):
    obj = {
        'image_id': image_id,
        'member': member_id,
    }
    obj.update(kwargs)
    return obj


def _db_task_fixture(task_id, **kwargs):
    default_datetime = timeutils.utcnow()
    obj = {
        'id': task_id,
        'status': kwargs.get('status', 'pending'),
        'type': 'import',
        'input': kwargs.get('input', {}),
        'result': None,
        'owner': None,
        'image_id': kwargs.get('image_id'),
        'user_id': kwargs.get('user_id'),
        'request_id': kwargs.get('request_id'),
        'message': None,
        'expires_at': default_datetime + datetime.timedelta(days=365),
        'created_at': default_datetime,
        'updated_at': default_datetime,
        'deleted_at': None,
        'deleted': False
    }
    obj.update(kwargs)
    return obj


def _db_node_reference_fixture(node_id, node_url, **kwargs):
    obj = {
        'node_reference_id': node_id,
        'node_reference_url': node_url,
    }
    obj.update(kwargs)
    return obj


def _db_cached_images_fixture(id, **kwargs):
    obj = {
        'id': id,
        'image_id': kwargs.get('image_id'),
        'size': kwargs.get('size'),
        'hits': kwargs.get('hits')
    }
    obj.update(kwargs)
    return obj


class TestImageRepo(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImageRepo, self).setUp()
        self.db = unit_test_utils.FakeDB(initialize=False)
        self.context = glance.context.RequestContext(
            user=USER1, tenant=TENANT1)
        self.image_repo = glance.db.ImageRepo(self.context, self.db)
        self.image_factory = glance.domain.ImageFactory()
        self._create_images()
        self._create_image_members()
        # Centralized cache
        self._create_node_references()
        self._create_cached_images()

    def _create_node_references(self):
        self.node_references = [
            _db_node_reference_fixture(NODE_REFERENCE_ID_1, 'node_url_1'),
            _db_node_reference_fixture(NODE_REFERENCE_ID_2, 'node_url_2'),
        ]
        [self.db.node_reference_create(
            None, node_reference['node_reference_url'],
            node_reference_id=node_reference['node_reference_id']
        ) for node_reference in self.node_references]

    def _create_cached_images(self):
        self.cached_images = [
            _db_cached_images_fixture(1, image_id=UUID1, size=256, hits=3),
            _db_cached_images_fixture(1, image_id=UUID3, size=1024, hits=0)
        ]
        [self.db.insert_cache_details(
            None, 'node_url_1', cached_image['image_id'],
            cached_image['size'], hits=cached_image['hits']
        ) for cached_image in self.cached_images]

    def _create_images(self):
        self.images = [
            _db_fixture(UUID1, owner=TENANT1, checksum=CHECKSUM,
                        name='1', size=256,
                        is_public=True, status='active',
                        locations=[{'url': UUID1_LOCATION,
                                    'metadata': UUID1_LOCATION_METADATA,
                                    'status': 'active'}]),
            _db_fixture(UUID2, owner=TENANT1, checksum=CHCKSUM1,
                        name='2', size=512, is_public=False),
            _db_fixture(UUID3, owner=TENANT3, checksum=CHCKSUM1,
                        name='3', size=1024, is_public=True,
                        locations=[{'url': UUID3_LOCATION,
                                    'metadata': {},
                                    'status': 'active'}]),
            _db_fixture(UUID4, owner=TENANT4, name='4', size=2048),
        ]
        [self.db.image_create(None, image) for image in self.images]

        # Create tasks associated with image
        self.tasks = [
            _db_task_fixture(
                TASK_ID_1, image_id=UUID1, status='completed',
                input={
                    "image_id": UUID1,
                    "import_req": {
                        "method": {
                            "name": "glance-direct"
                        },
                        "backend": ["fake-store"]
                    },
                },
                user_id=USER1,
                request_id='fake-request-id',
            ),
            _db_task_fixture(
                TASK_ID_2, image_id=UUID1, status='completed',
                input={
                    "image_id": UUID1,
                    "import_req": {
                        "method": {
                            "name": "copy-image"
                        },
                        "all_stores": True,
                        "all_stores_must_succeed": False,
                        "backend": ["fake-store", "fake_store_1"]
                    },
                },
                user_id=USER1,
                request_id='fake-request-id',
            ),
            _db_task_fixture(
                TASK_ID_3, status='completed',
                input={
                    "image_id": UUID2,
                    "import_req": {
                        "method": {
                            "name": "glance-direct"
                        },
                        "backend": ["fake-store"]
                    },
                },
            ),
        ]
        [self.db.task_create(None, task) for task in self.tasks]

        self.db.image_tag_set_all(None, UUID1, ['ping', 'pong'])

    def _create_image_members(self):
        self.image_members = [
            _db_image_member_fixture(UUID2, TENANT2),
            _db_image_member_fixture(UUID2, TENANT3, status='accepted'),
        ]
        [self.db.image_member_create(None, image_member)
            for image_member in self.image_members]

    def test_node_reference_get_by_url(self):
        node_reference = self.db.node_reference_get_by_url(self.context,
                                                           'node_url_1')
        self.assertEqual(NODE_REFERENCE_ID_1,
                         node_reference['node_reference_id'])

    def test_node_reference_get_by_url_not_found(self):
        self.assertRaises(exception.NotFound,
                          self.db.node_reference_get_by_url,
                          self.context,
                          'garbage_url')

    def test_get_cached_images(self):
        # Two images are cached on node 'node_url_1'
        cached_images = self.db.get_cached_images(self.context,
                                                  'node_url_1')
        self.assertEqual(2, len(cached_images))

        # Nothing is cached on node 'node_url_2'
        cached_images = self.db.get_cached_images(self.context,
                                                  'node_url_2')
        self.assertEqual(0, len(cached_images))

    def test_get_hit_count(self):
        # Hit count will be 3 for image UUID1
        self.assertEqual(3, self.db.get_hit_count(self.context,
                                                  UUID1, 'node_url_1'))

        # Hit count will be 0 for uncached image
        self.assertEqual(0, self.db.get_hit_count(self.context,
                                                  UUID2, 'node_url_1'))

    def test_delete_all_cached_images(self):
        # Verify that we have image cached
        cached_images = self.db.get_cached_images(self.context,
                                                  'node_url_1')
        self.assertEqual(2, len(cached_images))

        # Delete cached images from node_url_1
        self.db.delete_all_cached_images(self.context, 'node_url_1')

        # Verify that all cached images from node_url_1 are deleted
        cached_images = self.db.get_cached_images(self.context,
                                                  'node_url_1')
        self.assertEqual(0, len(cached_images))

    def test_delete_cached_image(self):
        # Verify that we have image cached
        cached_images = self.db.get_cached_images(self.context,
                                                  'node_url_1')
        self.assertEqual(2, len(cached_images))

        # Delete cached image from node_url_1
        self.db.delete_cached_image(self.context, UUID1, 'node_url_1')

        # Verify that given image from node_url_1 is deleted
        cached_images = self.db.get_cached_images(self.context,
                                                  'node_url_1')
        self.assertEqual(1, len(cached_images))

    def test_get_least_recently_accessed(self):
        recently_accessed = self.db.get_least_recently_accessed(
            self.context, 'node_url_1')
        # Verify we will only get one image in response
        self.assertEqual(UUID1, recently_accessed)

    def test_is_image_cached_for_node(self):
        # Verify UUID1 is cached for node_url_1
        self.assertTrue(self.db.is_image_cached_for_node(
            self.context, 'node_url_1', UUID1))

        # Verify UUID3 is not cached for node_url_2
        self.assertFalse(self.db.is_image_cached_for_node(
            self.context, 'node_url_2', UUID3))

    def test_update_hit_count(self):
        # Verify UUID1 on node_url_1 has 3 as hit count
        self.assertEqual(3, self.db.get_hit_count(self.context,
                                                  UUID1, 'node_url_1'))

        # Update the hit count of UUID1
        self.db.update_hit_count(self.context, UUID1, 'node_url_1')

        # Verify hit count is now 4
        self.assertEqual(4, self.db.get_hit_count(self.context,
                                                  UUID1, 'node_url_1'))

    def test_get(self):
        image = self.image_repo.get(UUID1)
        self.assertEqual(UUID1, image.image_id)
        self.assertEqual('1', image.name)
        self.assertEqual(set(['ping', 'pong']), image.tags)
        self.assertEqual('public', image.visibility)
        self.assertEqual('active', image.status)
        self.assertEqual(256, image.size)
        self.assertEqual(TENANT1, image.owner)

    def test_tasks_get_by_image(self):
        tasks = self.db.tasks_get_by_image(self.context, UUID1)
        self.assertEqual(2, len(tasks))
        for task in tasks:
            self.assertEqual(USER1, task['user_id'])
            self.assertEqual('fake-request-id', task['request_id'])
            self.assertEqual(UUID1, task['image_id'])

    def test_tasks_get_by_image_not_exists(self):
        tasks = self.db.tasks_get_by_image(self.context, UUID3)
        self.assertEqual(0, len(tasks))

    def test_location_value(self):
        image = self.image_repo.get(UUID3)
        self.assertEqual(UUID3_LOCATION, image.locations[0]['url'])

    def test_location_data_value(self):
        image = self.image_repo.get(UUID1)
        self.assertEqual(UUID1_LOCATION, image.locations[0]['url'])
        self.assertEqual(UUID1_LOCATION_METADATA,
                         image.locations[0]['metadata'])

    def test_location_data_exists(self):
        image = self.image_repo.get(UUID2)
        self.assertEqual([], image.locations)

    def test_get_not_found(self):
        fake_uuid = str(uuid.uuid4())
        exc = self.assertRaises(exception.ImageNotFound, self.image_repo.get,
                                fake_uuid)
        self.assertIn(fake_uuid, str(exc))

    def test_get_forbidden(self):
        self.assertRaises(exception.NotFound, self.image_repo.get, UUID4)

    def test_list(self):
        images = self.image_repo.list()
        image_ids = set([i.image_id for i in images])
        self.assertEqual(set([UUID1, UUID2, UUID3]), image_ids)

    def _do_test_list_status(self, status, expected):
        self.context = glance.context.RequestContext(
            user=USER1, tenant=TENANT3)
        self.image_repo = glance.db.ImageRepo(self.context, self.db)
        images = self.image_repo.list(member_status=status)
        self.assertEqual(expected, len(images))

    def test_list_status(self):
        self._do_test_list_status(None, 3)

    def test_list_status_pending(self):
        self._do_test_list_status('pending', 2)

    def test_list_status_rejected(self):
        self._do_test_list_status('rejected', 2)

    def test_list_status_all(self):
        self._do_test_list_status('all', 3)

    def test_list_with_marker(self):
        full_images = self.image_repo.list()
        full_ids = [i.image_id for i in full_images]
        marked_images = self.image_repo.list(marker=full_ids[0])
        actual_ids = [i.image_id for i in marked_images]
        self.assertEqual(full_ids[1:], actual_ids)

    def test_list_with_last_marker(self):
        images = self.image_repo.list()
        marked_images = self.image_repo.list(marker=images[-1].image_id)
        self.assertEqual(0, len(marked_images))

    def test_limited_list(self):
        limited_images = self.image_repo.list(limit=2)
        self.assertEqual(2, len(limited_images))

    def test_list_with_marker_and_limit(self):
        full_images = self.image_repo.list()
        full_ids = [i.image_id for i in full_images]
        marked_images = self.image_repo.list(marker=full_ids[0], limit=1)
        actual_ids = [i.image_id for i in marked_images]
        self.assertEqual(full_ids[1:2], actual_ids)

    def test_list_private_images(self):
        filters = {'visibility': 'private'}
        images = self.image_repo.list(filters=filters)
        self.assertEqual(0, len(images))

    def test_list_shared_images(self):
        filters = {'visibility': 'shared'}
        images = self.image_repo.list(filters=filters)
        image_ids = set([i.image_id for i in images])
        self.assertEqual(set([UUID2]), image_ids)

    def test_list_shared_images_other_tenant(self):
        # Create a private image owned by TENANT3
        image5 = _db_fixture(uuids.image5, owner=TENANT3,
                             name='5', size=512, is_public=False)
        self.db.image_create(None, image5)

        # Get a repo as TENANT3, since it has access to public,
        # shared, and private images
        context = glance.context.RequestContext(user=USER1, tenant=TENANT3)
        image_repo = glance.db.ImageRepo(context, self.db)
        images = {i.image_id: i for i in image_repo.list()}

        # No member set for public image UUID1
        self.assertIsNone(images[UUID1].member)

        # Member should be set to our tenant id for shared image UUID2
        self.assertEqual(TENANT3, images[UUID2].member)

        # No member set for private image5
        self.assertIsNone(images[uuids.image5].member)

    def test_list_all_images(self):
        filters = {'visibility': 'all'}
        images = self.image_repo.list(filters=filters)
        image_ids = set([i.image_id for i in images])
        self.assertEqual(set([UUID1, UUID2, UUID3]), image_ids)

    def test_list_with_checksum_filter_single_image(self):
        filters = {'checksum': CHECKSUM}
        images = self.image_repo.list(filters=filters)
        image_ids = list([i.image_id for i in images])
        self.assertEqual(1, len(image_ids))
        self.assertEqual([UUID1], image_ids)

    def test_list_with_checksum_filter_multiple_images(self):
        filters = {'checksum': CHCKSUM1}
        images = self.image_repo.list(filters=filters)
        image_ids = list([i.image_id for i in images])
        self.assertEqual(2, len(image_ids))
        self.assertIn(UUID2, image_ids)
        self.assertIn(UUID3, image_ids)

    def test_list_with_wrong_checksum(self):
        WRONG_CHKSUM = 'd2fd42f979e1ed1aafadc7eb9354bff839c858cd'
        filters = {'checksum': WRONG_CHKSUM}
        images = self.image_repo.list(filters=filters)
        self.assertEqual(0, len(images))

    def test_list_with_tags_filter_single_tag(self):
        filters = {'tags': ['ping']}
        images = self.image_repo.list(filters=filters)
        image_ids = list([i.image_id for i in images])
        self.assertEqual(1, len(image_ids))
        self.assertEqual([UUID1], image_ids)

    def test_list_with_tags_filter_multiple_tags(self):
        filters = {'tags': ['ping', 'pong']}
        images = self.image_repo.list(filters=filters)
        image_ids = list([i.image_id for i in images])
        self.assertEqual(1, len(image_ids))
        self.assertEqual([UUID1], image_ids)

    def test_list_with_tags_filter_multiple_tags_and_nonexistent(self):
        filters = {'tags': ['ping', 'fake']}
        images = self.image_repo.list(filters=filters)
        image_ids = list([i.image_id for i in images])
        self.assertEqual(0, len(image_ids))

    def test_list_with_wrong_tags(self):
        filters = {'tags': ['fake']}
        images = self.image_repo.list(filters=filters)
        self.assertEqual(0, len(images))

    def test_list_public_images(self):
        filters = {'visibility': 'public'}
        images = self.image_repo.list(filters=filters)
        image_ids = set([i.image_id for i in images])
        self.assertEqual(set([UUID1, UUID3]), image_ids)

    def test_sorted_list(self):
        images = self.image_repo.list(sort_key=['size'], sort_dir=['asc'])
        image_ids = [i.image_id for i in images]
        self.assertEqual([UUID1, UUID2, UUID3], image_ids)

    def test_sorted_list_with_multiple_keys(self):
        temp_id = 'd80a1a6c-bd1f-41c5-90ee-81afedb1d58d'
        image = _db_fixture(temp_id, owner=TENANT1, checksum=CHECKSUM,
                            name='1', size=1024,
                            is_public=True, status='active',
                            locations=[{'url': UUID1_LOCATION,
                                        'metadata': UUID1_LOCATION_METADATA,
                                        'status': 'active'}])
        self.db.image_create(None, image)
        images = self.image_repo.list(sort_key=['name', 'size'],
                                      sort_dir=['asc'])
        image_ids = [i.image_id for i in images]
        self.assertEqual([UUID1, temp_id, UUID2, UUID3], image_ids)

        images = self.image_repo.list(sort_key=['size', 'name'],
                                      sort_dir=['asc'])
        image_ids = [i.image_id for i in images]
        self.assertEqual([UUID1, UUID2, temp_id, UUID3], image_ids)

    def test_sorted_list_with_multiple_dirs(self):
        temp_id = 'd80a1a6c-bd1f-41c5-90ee-81afedb1d58d'
        image = _db_fixture(temp_id, owner=TENANT1, checksum=CHECKSUM,
                            name='1', size=1024,
                            is_public=True, status='active',
                            locations=[{'url': UUID1_LOCATION,
                                        'metadata': UUID1_LOCATION_METADATA,
                                        'status': 'active'}])
        self.db.image_create(None, image)
        images = self.image_repo.list(sort_key=['name', 'size'],
                                      sort_dir=['asc', 'desc'])
        image_ids = [i.image_id for i in images]
        self.assertEqual([temp_id, UUID1, UUID2, UUID3], image_ids)

        images = self.image_repo.list(sort_key=['name', 'size'],
                                      sort_dir=['desc', 'asc'])
        image_ids = [i.image_id for i in images]
        self.assertEqual([UUID3, UUID2, UUID1, temp_id], image_ids)

    def test_add_image(self):
        image = self.image_factory.new_image(name='added image')
        self.assertEqual(image.updated_at, image.created_at)
        self.image_repo.add(image)
        retreived_image = self.image_repo.get(image.image_id)
        self.assertEqual('added image', retreived_image.name)
        self.assertEqual(image.updated_at, retreived_image.updated_at)

    def test_save_image(self):
        image = self.image_repo.get(UUID1)
        original_update_time = image.updated_at
        image.name = 'foo'
        image.tags = ['king', 'kong']
        self.image_repo.save(image)
        current_update_time = image.updated_at
        self.assertGreater(current_update_time, original_update_time)
        image = self.image_repo.get(UUID1)
        self.assertEqual('foo', image.name)
        self.assertEqual(set(['king', 'kong']), image.tags)
        self.assertEqual(current_update_time, image.updated_at)

    def test_save_image_not_found(self):
        fake_uuid = str(uuid.uuid4())
        image = self.image_repo.get(UUID1)
        image.image_id = fake_uuid
        exc = self.assertRaises(exception.ImageNotFound, self.image_repo.save,
                                image)
        self.assertIn(fake_uuid, str(exc))

    def test_save_excludes_atomic_props(self):
        fake_uuid = str(uuid.uuid4())
        image = self.image_repo.get(UUID1)

        # Try to set the property normally
        image.extra_properties['os_glance_import_task'] = fake_uuid
        self.image_repo.save(image)

        # Expect it was ignored
        image = self.image_repo.get(UUID1)
        self.assertNotIn('os_glance_import_task', image.extra_properties)

        # Set the property atomically
        self.image_repo.set_property_atomic(image,
                                            'os_glance_import_task', fake_uuid)
        # Expect it is set
        image = self.image_repo.get(UUID1)
        self.assertEqual(fake_uuid,
                         image.extra_properties['os_glance_import_task'])

        # Try to clobber it
        image.extra_properties['os_glance_import_task'] = 'foo'
        self.image_repo.save(image)

        # Expect it is unchanged
        image = self.image_repo.get(UUID1)
        self.assertEqual(fake_uuid,
                         image.extra_properties['os_glance_import_task'])

        # Try to delete it
        del image.extra_properties['os_glance_import_task']
        self.image_repo.save(image)

        # Expect it is still present and set accordingly
        image = self.image_repo.get(UUID1)
        self.assertEqual(fake_uuid,
                         image.extra_properties['os_glance_import_task'])

    def test_remove_image(self):
        image = self.image_repo.get(UUID1)
        previous_update_time = image.updated_at
        self.image_repo.remove(image)
        self.assertGreater(image.updated_at, previous_update_time)
        self.assertRaises(exception.ImageNotFound, self.image_repo.get, UUID1)

    def test_remove_image_not_found(self):
        fake_uuid = str(uuid.uuid4())
        image = self.image_repo.get(UUID1)
        image.image_id = fake_uuid
        exc = self.assertRaises(
            exception.ImageNotFound, self.image_repo.remove, image)
        self.assertIn(fake_uuid, str(exc))

    def test_restore_image_status(self):
        image_id = uuid.uuid4()
        image = _db_fixture(image_id, name='restore_test', size=256,
                            is_public=True, status='pending_delete')
        self.db.image_create(self.context, image)
        self.db.image_restore(self.context, image_id)
        image = self.db.image_get(self.context, image_id)
        self.assertEqual(image['status'], 'active')

    def test_restore_image_status_not_found(self):
        image_id = uuid.uuid4()
        self.assertRaises(exception.ImageNotFound,
                          self.db.image_restore,
                          self.context,
                          image_id)

    def test_restore_image_status_not_pending_delete(self):
        image_id = uuid.uuid4()
        image = _db_fixture(image_id, name='restore_test', size=256,
                            is_public=True, status='deleted')
        self.db.image_create(self.context, image)
        self.assertRaises(exception.Conflict,
                          self.db.image_restore,
                          self.context,
                          image_id)

    def test_image_set_property_atomic(self):
        image_id = uuid.uuid4()
        image = _db_fixture(image_id, name='test')

        self.assertRaises(exception.ImageNotFound,
                          self.db.image_set_property_atomic,
                          image_id, 'foo', 'bar')

        self.db.image_create(self.context, image)
        self.db.image_set_property_atomic(image_id, 'foo', 'bar')
        image = self.db.image_get(self.context, image_id)
        self.assertEqual('foo', image['properties'][0]['name'])
        self.assertEqual('bar', image['properties'][0]['value'])

    def test_set_property_atomic(self):
        image = self.image_repo.get(UUID1)
        self.image_repo.set_property_atomic(image, 'foo', 'bar')
        image = self.image_repo.get(image.image_id)
        self.assertEqual({'foo': 'bar'}, image.extra_properties)

    def test_image_delete_property_atomic(self):
        image_id = uuid.uuid4()
        image = _db_fixture(image_id, name='test')

        self.assertRaises(exception.NotFound,
                          self.db.image_delete_property_atomic,
                          image_id, 'foo', 'bar')
        self.db.image_create(self.context, image)
        self.db.image_set_property_atomic(image_id, 'foo', 'bar')
        self.db.image_delete_property_atomic(image_id, 'foo', 'bar')
        image = self.image_repo.get(image_id)
        self.assertEqual({}, image.extra_properties)

    def test_delete_property_atomic(self):
        image = self.image_repo.get(UUID1)
        self.image_repo.set_property_atomic(image, 'foo', 'bar')
        image = self.image_repo.get(image.image_id)
        self.image_repo.delete_property_atomic(image, 'foo', 'bar')
        image = self.image_repo.get(image.image_id)
        self.assertEqual({}, image.extra_properties)


class TestEncryptedLocations(test_utils.BaseTestCase):
    def setUp(self):
        super(TestEncryptedLocations, self).setUp()
        self.db = unit_test_utils.FakeDB(initialize=False)
        self.context = glance.context.RequestContext(
            user=USER1, tenant=TENANT1)
        self.image_repo = glance.db.ImageRepo(self.context, self.db)
        self.image_factory = glance.domain.ImageFactory()
        self.crypt_key = '0123456789abcdef'
        self.config(metadata_encryption_key=self.crypt_key)
        self.foo_bar_location = [{'url': 'foo', 'metadata': {},
                                  'status': 'active'},
                                 {'url': 'bar', 'metadata': {},
                                  'status': 'active'}]

    def test_encrypt_locations_on_add(self):
        image = self.image_factory.new_image(UUID1)
        image.locations = self.foo_bar_location
        self.image_repo.add(image)
        db_data = self.db.image_get(self.context, UUID1)
        self.assertNotEqual(db_data['locations'], ['foo', 'bar'])
        decrypted_locations = [crypt.urlsafe_decrypt(self.crypt_key,
                                                     location['url'])
                               for location in db_data['locations']]
        self.assertEqual([location['url']
                          for location in self.foo_bar_location],
                         decrypted_locations)

    def test_encrypt_locations_on_save(self):
        image = self.image_factory.new_image(UUID1)
        self.image_repo.add(image)
        image.locations = self.foo_bar_location
        self.image_repo.save(image)
        db_data = self.db.image_get(self.context, UUID1)
        self.assertNotEqual(db_data['locations'], ['foo', 'bar'])
        decrypted_locations = [crypt.urlsafe_decrypt(self.crypt_key,
                                                     location['url'])
                               for location in db_data['locations']]
        self.assertEqual([location['url']
                          for location in self.foo_bar_location],
                         decrypted_locations)

    def test_decrypt_locations_on_get(self):
        url_loc = ['ping', 'pong']
        orig_locations = [{'url': location, 'metadata': {}, 'status': 'active'}
                          for location in url_loc]
        encrypted_locs = [crypt.urlsafe_encrypt(self.crypt_key, location)
                          for location in url_loc]
        encrypted_locations = [{'url': location,
                                'metadata': {},
                                'status': 'active'}
                               for location in encrypted_locs]
        self.assertNotEqual(encrypted_locations, orig_locations)
        db_data = _db_fixture(UUID1, owner=TENANT1,
                              locations=encrypted_locations)
        self.db.image_create(None, db_data)
        image = self.image_repo.get(UUID1)
        self.assertIn('id', image.locations[0])
        self.assertIn('id', image.locations[1])
        image.locations[0].pop('id')
        image.locations[1].pop('id')
        self.assertEqual(orig_locations, image.locations)

    def test_decrypt_locations_on_list(self):
        url_loc = ['ping', 'pong']
        orig_locations = [{'url': location, 'metadata': {}, 'status': 'active'}
                          for location in url_loc]
        encrypted_locs = [crypt.urlsafe_encrypt(self.crypt_key, location)
                          for location in url_loc]
        encrypted_locations = [{'url': location,
                                'metadata': {},
                                'status': 'active'}
                               for location in encrypted_locs]
        self.assertNotEqual(encrypted_locations, orig_locations)
        db_data = _db_fixture(UUID1, owner=TENANT1,
                              locations=encrypted_locations)
        self.db.image_create(None, db_data)
        image = self.image_repo.list()[0]
        self.assertIn('id', image.locations[0])
        self.assertIn('id', image.locations[1])
        image.locations[0].pop('id')
        image.locations[1].pop('id')
        self.assertEqual(orig_locations, image.locations)


class TestImageMemberRepo(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImageMemberRepo, self).setUp()
        self.db = unit_test_utils.FakeDB(initialize=False)
        self.context = glance.context.RequestContext(
            user=USER1, tenant=TENANT1)
        self.image_repo = glance.db.ImageRepo(self.context, self.db)
        self.image_member_factory = glance.domain.ImageMemberFactory()
        self._create_images()
        self._create_image_members()
        image = self.image_repo.get(UUID1)
        self.image_member_repo = glance.db.ImageMemberRepo(self.context,
                                                           self.db, image)

    def _create_images(self):
        self.images = [
            _db_fixture(UUID1, owner=TENANT1, name='1', size=256,
                        status='active'),
            _db_fixture(UUID2, owner=TENANT1, name='2',
                        size=512, visibility='shared'),
        ]
        [self.db.image_create(None, image) for image in self.images]

        self.db.image_tag_set_all(None, UUID1, ['ping', 'pong'])

    def _create_image_members(self):
        self.image_members = [
            _db_image_member_fixture(UUID1, TENANT2),
            _db_image_member_fixture(UUID1, TENANT3),
        ]
        [self.db.image_member_create(None, image_member)
            for image_member in self.image_members]

    def test_list(self):
        image_members = self.image_member_repo.list()
        image_member_ids = set([i.member_id for i in image_members])
        self.assertEqual(set([TENANT2, TENANT3]), image_member_ids)

    def test_list_no_members(self):
        image = self.image_repo.get(UUID2)
        self.image_member_repo_uuid2 = glance.db.ImageMemberRepo(
            self.context, self.db, image)
        image_members = self.image_member_repo_uuid2.list()
        image_member_ids = set([i.member_id for i in image_members])
        self.assertEqual(set([]), image_member_ids)

    def test_save_image_member(self):
        image_member = self.image_member_repo.get(TENANT2)
        image_member.status = 'accepted'
        self.image_member_repo.save(image_member)
        image_member_updated = self.image_member_repo.get(TENANT2)
        self.assertEqual(image_member.id, image_member_updated.id)
        self.assertEqual('accepted', image_member_updated.status)

    def test_add_image_member(self):
        image = self.image_repo.get(UUID1)
        image_member = self.image_member_factory.new_image_member(image,
                                                                  TENANT4)
        self.assertIsNone(image_member.id)
        self.image_member_repo.add(image_member)
        retreived_image_member = self.image_member_repo.get(TENANT4)
        self.assertIsNotNone(retreived_image_member.id)
        self.assertEqual(image_member.image_id,
                         retreived_image_member.image_id)
        self.assertEqual(image_member.member_id,
                         retreived_image_member.member_id)
        self.assertEqual('pending', retreived_image_member.status)

    def test_add_duplicate_image_member(self):
        image = self.image_repo.get(UUID1)
        image_member = self.image_member_factory.new_image_member(image,
                                                                  TENANT4)
        self.assertIsNone(image_member.id)
        self.image_member_repo.add(image_member)
        retreived_image_member = self.image_member_repo.get(TENANT4)
        self.assertIsNotNone(retreived_image_member.id)
        self.assertEqual(image_member.image_id,
                         retreived_image_member.image_id)
        self.assertEqual(image_member.member_id,
                         retreived_image_member.member_id)
        self.assertEqual('pending', retreived_image_member.status)

        self.assertRaises(exception.Duplicate, self.image_member_repo.add,
                          image_member)

    def test_get_image_member(self):
        image = self.image_repo.get(UUID1)
        image_member = self.image_member_factory.new_image_member(image,
                                                                  TENANT4)
        self.assertIsNone(image_member.id)
        self.image_member_repo.add(image_member)

        member = self.image_member_repo.get(image_member.member_id)

        self.assertEqual(member.id, image_member.id)
        self.assertEqual(member.image_id, image_member.image_id)
        self.assertEqual(member.member_id, image_member.member_id)
        self.assertEqual('pending', member.status)

    def test_get_nonexistent_image_member(self):
        fake_image_member_id = 'fake'
        self.assertRaises(exception.NotFound, self.image_member_repo.get,
                          fake_image_member_id)

    def test_remove_image_member(self):
        image_member = self.image_member_repo.get(TENANT2)
        self.image_member_repo.remove(image_member)
        self.assertRaises(exception.NotFound, self.image_member_repo.get,
                          TENANT2)

    def test_remove_image_member_does_not_exist(self):
        fake_uuid = str(uuid.uuid4())
        image = self.image_repo.get(UUID2)
        fake_member = glance.domain.ImageMemberFactory().new_image_member(
            image, TENANT4)
        fake_member.id = fake_uuid
        exc = self.assertRaises(exception.NotFound,
                                self.image_member_repo.remove,
                                fake_member)
        self.assertIn(fake_uuid, str(exc))


class TestTaskRepo(test_utils.BaseTestCase):

    def setUp(self):
        super(TestTaskRepo, self).setUp()
        self.db = unit_test_utils.FakeDB(initialize=False)
        self.context = glance.context.RequestContext(user=USER1,
                                                     tenant=TENANT1)
        self.task_repo = glance.db.TaskRepo(self.context, self.db)
        self.task_factory = glance.domain.TaskFactory()
        self.fake_task_input = ('{"import_from": '
                                '"swift://cloud.foo/account/mycontainer/path"'
                                ',"import_from_format": "qcow2"}')
        self._create_tasks()

    def _create_tasks(self):
        self.tasks = [
            _db_task_fixture(UUID1, type='import', status='pending',
                             input=self.fake_task_input,
                             result='',
                             owner=TENANT1,
                             message='',
                             ),
            _db_task_fixture(UUID2, type='import', status='processing',
                             input=self.fake_task_input,
                             result='',
                             owner=TENANT1,
                             message='',
                             ),
            _db_task_fixture(UUID3, type='import', status='failure',
                             input=self.fake_task_input,
                             result='',
                             owner=TENANT1,
                             message='',
                             ),
            _db_task_fixture(UUID4, type='import', status='success',
                             input=self.fake_task_input,
                             result='',
                             owner=TENANT2,
                             message='',
                             ),
        ]
        [self.db.task_create(None, task) for task in self.tasks]

    def test_get(self):
        task = self.task_repo.get(UUID1)
        self.assertEqual(task.task_id, UUID1)
        self.assertEqual('import', task.type)
        self.assertEqual('pending', task.status)
        self.assertEqual(task.task_input, self.fake_task_input)
        self.assertEqual('', task.result)
        self.assertEqual('', task.message)
        self.assertEqual(task.owner, TENANT1)

    def test_get_not_found(self):
        self.assertRaises(exception.NotFound,
                          self.task_repo.get,
                          str(uuid.uuid4()))

    def test_get_forbidden(self):
        self.assertRaises(exception.NotFound,
                          self.task_repo.get,
                          UUID4)

    def test_list(self):
        tasks = self.task_repo.list()
        task_ids = set([i.task_id for i in tasks])
        self.assertEqual(set([UUID1, UUID2, UUID3]), task_ids)

    def test_list_with_type(self):
        filters = {'type': 'import'}
        tasks = self.task_repo.list(filters=filters)
        task_ids = set([i.task_id for i in tasks])
        self.assertEqual(set([UUID1, UUID2, UUID3]), task_ids)

    def test_list_with_status(self):
        filters = {'status': 'failure'}
        tasks = self.task_repo.list(filters=filters)
        task_ids = set([i.task_id for i in tasks])
        self.assertEqual(set([UUID3]), task_ids)

    def test_list_with_marker(self):
        full_tasks = self.task_repo.list()
        full_ids = [i.task_id for i in full_tasks]
        marked_tasks = self.task_repo.list(marker=full_ids[0])
        actual_ids = [i.task_id for i in marked_tasks]
        self.assertEqual(full_ids[1:], actual_ids)

    def test_list_with_last_marker(self):
        tasks = self.task_repo.list()
        marked_tasks = self.task_repo.list(marker=tasks[-1].task_id)
        self.assertEqual(0, len(marked_tasks))

    def test_limited_list(self):
        limited_tasks = self.task_repo.list(limit=2)
        self.assertEqual(2, len(limited_tasks))

    def test_list_with_marker_and_limit(self):
        full_tasks = self.task_repo.list()
        full_ids = [i.task_id for i in full_tasks]
        marked_tasks = self.task_repo.list(marker=full_ids[0], limit=1)
        actual_ids = [i.task_id for i in marked_tasks]
        self.assertEqual(full_ids[1:2], actual_ids)

    def test_sorted_list(self):
        tasks = self.task_repo.list(sort_key='status', sort_dir='desc')
        task_ids = [i.task_id for i in tasks]
        self.assertEqual([UUID2, UUID1, UUID3], task_ids)

    def test_add_task(self):
        task_type = 'import'
        image_id = 'fake_image_id'
        user_id = 'fake_user'
        request_id = 'fake_request_id'
        task = self.task_factory.new_task(task_type, None, image_id,
                                          user_id, request_id,
                                          task_input=self.fake_task_input)
        self.assertEqual(task.updated_at, task.created_at)
        self.task_repo.add(task)
        retrieved_task = self.task_repo.get(task.task_id)
        self.assertEqual(task.updated_at, retrieved_task.updated_at)
        self.assertEqual(self.fake_task_input, retrieved_task.task_input)
        self.assertEqual(image_id, task.image_id)
        self.assertEqual(user_id, task.user_id)
        self.assertEqual(request_id, task.request_id)

    def test_save_task(self):
        task = self.task_repo.get(UUID1)
        original_update_time = task.updated_at
        self.task_repo.save(task)
        current_update_time = task.updated_at
        self.assertGreater(current_update_time, original_update_time)
        task = self.task_repo.get(UUID1)
        self.assertEqual(current_update_time, task.updated_at)

    def test_remove_task(self):
        task = self.task_repo.get(UUID1)
        self.task_repo.remove(task)
        self.assertRaises(exception.NotFound,
                          self.task_repo.get,
                          task.task_id)


class TestImageDeleteRace(test_utils.BaseTestCase):

    @mock.patch.object(api, 'session_for_write')
    @mock.patch.object(api, 'LOG')
    def test_image_property_delete_stale_data(
        self, mock_LOG, mock_session_for,
    ):
        mock_context = mock.MagicMock()
        mock_session = mock_session_for.return_value
        mock_result = (mock_session.__enter__.return_value.
                       query.return_value.
                       filter_by.return_value.
                       one.return_value)
        mock_result.delete.side_effect = sa_orm.exc.StaleDataError('myerror')
        # StaleDataError should not be raised
        r = api.image_property_delete(mock_context, 'myprop', 'myimage')
        # We should not get the property back
        self.assertIsNone(r)
        # Make sure we logged it
        mock_LOG.debug.assert_called_once_with(
            'StaleDataError while deleting property %(prop)r '
            'from image %(image)r likely means we raced during delete: '
            '%(err)s', {'prop': 'myprop', 'image': 'myimage',
                        'err': 'myerror'})

    @mock.patch.object(api, 'session_for_write')
    def test_image_property_delete_exception(self, mock_session_for):
        mock_context = mock.MagicMock()
        mock_session = mock_session_for.return_value
        mock_result = (mock_session.__enter__.return_value.
                       query.return_value.
                       filter_by.return_value.
                       one.return_value)
        mock_result.delete.side_effect = RuntimeError
        # Any other exception should be raised
        self.assertRaises(RuntimeError,
                          api.image_property_delete,
                          mock_context, 'myprop', 'myimage')



# Copyright 2018 RedHat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
import json
import os
from unittest import mock

import glance_store
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_utils.imageutils import format_inspector

import glance.async_.flows.api_image_import as import_flow
import glance.async_.flows.plugins.image_conversion as image_conversion
from glance.async_ import utils as async_utils
from glance.common import utils
from glance import domain
from glance import gateway
import glance.tests.utils as test_utils

CONF = cfg.CONF


UUID1 = 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d'
TENANT1 = '6838eb7b-6ded-434a-882c-b344c77fe8df'


class TestConvertImageTask(test_utils.BaseTestCase):

    def setUp(self):
        super(TestConvertImageTask, self).setUp()

        glance_store.register_opts(CONF)
        self.config(default_store='file',
                    stores=['file', 'http'],
                    filesystem_store_datadir=self.test_dir,
                    group="glance_store")
        self.config(output_format='qcow2',
                    group='image_conversion')
        glance_store.create_stores(CONF)

        self.work_dir = os.path.join(self.test_dir, 'work_dir')
        utils.safe_mkdirs(self.work_dir)
        self.config(work_dir=self.work_dir, group='task')

        self.context = mock.MagicMock()
        self.img_repo = mock.MagicMock()
        self.task_repo = mock.MagicMock()
        self.stores = mock.MagicMock()
        self.image_id = UUID1

        self.gateway = gateway.Gateway()
        self.task_factory = domain.TaskFactory()
        self.img_factory = self.gateway.get_image_factory(self.context)
        self.image = self.img_factory.new_image(image_id=self.image_id,
                                                disk_format='raw',
                                                container_format='bare')

        task_input = {
            "import_from": "http://cloud.foo/image.raw",
            "import_from_format": "raw",
            "image_properties": {'disk_format': 'raw',
                                 'container_format': 'bare'}
        }

        task_ttl = CONF.task.task_time_to_live

        self.task_type = 'import'
        request_id = 'fake_request_id'
        user_id = 'fake_user'
        self.task = self.task_factory.new_task(self.task_type, TENANT1,
                                               self.image_id, user_id,
                                               request_id,
                                               task_time_to_live=task_ttl,
                                               task_input=task_input)

        self.image.extra_properties = {
            'os_glance_import_task': self.task.task_id,
            'os_glance_importing_to_stores': mock.MagicMock(),
            'os_glance_failed_import': ""
        }
        self.wrapper = import_flow.ImportActionWrapper(self.img_repo,
                                                       self.image_id,
                                                       self.task.task_id)

        self.detect_file_format_mock = mock.MagicMock()
        self.useFixture(fixtures.MockPatch(
            'oslo_utils.imageutils.format_inspector.detect_file_format',
            self.detect_file_format_mock))

    @mock.patch.object(os, 'stat')
    @mock.patch.object(os, 'remove')
    def test_image_convert_success(self, mock_os_remove, mock_os_stat):
        mock_os_remove.return_value = None
        mock_os_stat.return_value.st_size = 123
        image_convert = image_conversion._ConvertImage(self.context,
                                                       self.task.task_id,
                                                       self.task_type,
                                                       self.wrapper,
                                                       self.stores)

        self.task_repo.get.return_value = self.task
        image = mock.MagicMock(image_id=self.image_id, virtual_size=None,
                               extra_properties={
                                   'os_glance_import_task': self.task.task_id},
                               disk_format='raw')
        self.img_repo.get.return_value = image

        with mock.patch.object(processutils, 'execute') as exc_mock:
            exc_mock.return_value = ("", None)
            with mock.patch.object(json, 'loads') as jloads_mock:
                jloads_mock.return_value = {'format': 'raw',
                                            'virtual-size': 456}
                inspector = self.detect_file_format_mock.return_value
                inspector.__str__.side_effect = ['raw', 'qcow2']
                image_convert.execute('file:///test/path.raw')

                # NOTE(hemanthm): Asserting that the source format is passed
                # to qemu-utis to avoid inferring the image format. This
                # shields us from an attack vector described at
                # https://bugs.launchpad.net/glance/+bug/1449062/comments/72
                self.assertIn('-f', exc_mock.call_args[0])
                self.assertEqual("qcow2", image.disk_format)

        self.assertEqual('bare', image.container_format)
        self.assertEqual('qcow2', image.disk_format)
        self.assertEqual(456, image.virtual_size)
        self.assertEqual(123, image.size)

    @mock.patch('os.remove')
    @mock.patch('os.stat')
    @mock.patch('oslo_concurrency.processutils.trycmd')
    @mock.patch('glance.async_.flows.plugins.image_conversion.LOG')
    def test_image_convert_excluded_for_iso(self, mock_log, mock_run,
                                            mock_stat, mock_remove, fmt='iso'):
        mock_stat.return_value = mock.MagicMock(stsize=123)
        stdout = json.dumps({'format': 'raw'})
        mock_run.return_value = (stdout, '')
        image_convert = image_conversion._ConvertImage(self.context,
                                                       self.task.task_id,
                                                       self.task_type,
                                                       self.wrapper,
                                                       self.stores)
        self.task_repo.get.return_value = self.task
        image = mock.MagicMock(image_id=self.image_id, virtual_size=None,
                               extra_properties={
                                   'os_glance_import_task': self.task.task_id},
                               disk_format=fmt)

        self.img_repo.get.return_value = image
        inspector = self.detect_file_format_mock.return_value
        inspector.__str__.return_value = 'iso'

        image_convert.execute('file:///test/path.iso')
        self.assertEqual(fmt, image.disk_format)
        mock_log.debug.assert_called_once_with(
            "Avoiding conversion of an image %s having "
            "`iso` disk format.", self.image_id)

    def test_image_convert_iso_lie_raw(self):
        e = self.assertRaises(
            RuntimeError, self.test_image_convert_excluded_for_iso,
            fmt='raw')
        self.assertEqual('Image has disallowed configuration',
                         str(e))

    def test_image_convert_iso_lie_qcow2(self):
        e = self.assertRaises(
            RuntimeError, self.test_image_convert_excluded_for_iso,
            fmt='qcow2')
        self.assertEqual('Image has disallowed configuration',
                         str(e))

    def _setup_image_convert_info_fail(self, disk_format='qcow2'):
        image_convert = image_conversion._ConvertImage(self.context,
                                                       self.task.task_id,
                                                       self.task_type,
                                                       self.wrapper,
                                                       self.stores)

        self.task_repo.get.return_value = self.task
        image = mock.MagicMock(image_id=self.image_id, virtual_size=None,
                               extra_properties={
                                   'os_glance_import_task': self.task.task_id},
                               disk_format=disk_format)
        self.img_repo.get.return_value = image
        return image_convert

    def test_image_convert_fails_inspection(self):
        convert = self._setup_image_convert_info_fail()
        with mock.patch.object(processutils, 'execute') as exc_mock:
            inspector = self.detect_file_format_mock.return_value
            inspector.__str__.return_value = 'qcow2'
            exc_mock.side_effect = OSError('fail')
            self.assertRaises(OSError,
                              convert.execute, 'file:///test/path.raw')
            exc_mock.assert_called_once_with(
                'qemu-img', 'info',
                '-f', 'qcow2',
                '--output=json',
                '/test/path.raw',
                prlimit=async_utils.QEMU_IMG_PROC_LIMITS,
                python_exec=convert.python,
                log_errors=processutils.LOG_ALL_ERRORS)
        # Make sure we did not update the image
        self.img_repo.save.assert_not_called()

    def test_image_convert_inspection_reports_error(self):
        convert = self._setup_image_convert_info_fail()
        with mock.patch.object(processutils, 'execute') as exc_mock:
            exc_mock.return_value = '', 'some error'
            inspector = self.detect_file_format_mock.return_value
            inspector.__str__.return_value = 'qcow2'
            self.assertRaises(RuntimeError,
                              convert.execute, 'file:///test/path.raw')
            exc_mock.assert_called_once_with(
                'qemu-img', 'info',
                '-f', 'qcow2',
                '--output=json',
                '/test/path.raw',
                prlimit=async_utils.QEMU_IMG_PROC_LIMITS,
                python_exec=convert.python,
                log_errors=processutils.LOG_ALL_ERRORS)
        # Make sure we did not update the image
        self.img_repo.save.assert_not_called()

    def test_image_convert_invalid_qcow(self):
        data = {'format': 'qcow2',
                'backing-filename': '/etc/hosts'}

        convert = self._setup_image_convert_info_fail()
        with mock.patch.object(processutils, 'execute') as exc_mock:
            inspector = self.detect_file_format_mock.return_value
            inspector.__str__.return_value = 'qcow2'
            exc_mock.return_value = json.dumps(data), ''
            e = self.assertRaises(RuntimeError,
                                  convert.execute, 'file:///test/path.qcow')
            self.assertEqual('QCOW images with backing files are not allowed',
                             str(e))

    def test_image_convert_invalid_qcow_data_file(self):
        data = {'format': 'qcow2',
                'format-specific': {
                    'data': {
                        'data-file': '/etc/hosts',
                    },
                }}

        convert = self._setup_image_convert_info_fail()
        with mock.patch.object(processutils, 'execute') as exc_mock:
            exc_mock.return_value = json.dumps(data), ''
            inspector = self.detect_file_format_mock.return_value
            inspector.__str__.return_value = 'qcow2'
            e = self.assertRaises(RuntimeError,
                                  convert.execute, 'file:///test/path.qcow')
            self.assertEqual('QCOW images with data-file set are not allowed',
                             str(e))

    def test_image_convert_no_inspector_match(self):
        convert = self._setup_image_convert_info_fail()
        self.assertRaisesRegex(RuntimeError,
                               'Image format mismatch',
                               convert.execute, 'file:///test/path.hpfs')

    def test_image_convert_fails_inspection_safety_check(self):
        convert = self._setup_image_convert_info_fail()
        inspector = self.detect_file_format_mock.return_value
        inspector.safety_check.side_effect = (
            format_inspector.SafetyCheckFailed({}))
        self.assertRaisesRegex(RuntimeError,
                               'Image has disallowed configuration',
                               convert.execute, 'file:///test/path.qcow')

    def test_image_convert_fails_inspection_format_check(self):
        convert = self._setup_image_convert_info_fail()
        self.detect_file_format_mock.side_effect = (
            format_inspector.ImageFormatError())
        self.assertRaisesRegex(RuntimeError,
                               'Image format detection failed',
                               convert.execute, 'file:///test/path.qcow')

    def test_image_convert_fails_inspection_error(self):
        convert = self._setup_image_convert_info_fail()
        self.detect_file_format_mock.side_effect = ValueError
        self.assertRaisesRegex(RuntimeError,
                               'Unable to inspect image',
                               convert.execute, 'file:///test/path.qcow')

    def _test_image_convert_invalid_vmdk(self):
        data = {'format': 'vmdk',
                'format-specific': {
                    'data': {
                        'create-type': 'monolithicFlat',
                    }}}

        convert = self._setup_image_convert_info_fail(disk_format='vmdk')
        with mock.patch.object(processutils, 'execute') as exc_mock:
            inspector = self.detect_file_format_mock.return_value
            inspector.__str__.side_effect = ['vmdk', 'qcow2']
            exc_mock.return_value = json.dumps(data), ''
            convert.execute('file:///test/path.vmdk')

    def test_image_convert_invalid_vmdk(self):
        e = self.assertRaises(RuntimeError,
                              self._test_image_convert_invalid_vmdk)
        self.assertEqual('Invalid VMDK create-type specified', str(e))

    def test_image_convert_valid_vmdk_no_types(self):
        with mock.patch.object(CONF.image_format, 'vmdk_allowed_types',
                               new=[]):
            # We make it past the VMDK check and fail because our file
            # does not exist
            e = self.assertRaises(RuntimeError,
                                  self._test_image_convert_invalid_vmdk)
            self.assertEqual('Image is a VMDK, but no VMDK createType is '
                             'specified', str(e))

    def test_image_convert_valid_vmdk(self):
        with mock.patch.object(CONF.image_format, 'vmdk_allowed_types',
                               new=['monolithicSparse', 'monolithicFlat']):
            # We make it past the VMDK check and fail because our file
            # does not exist
            self.assertRaises(FileNotFoundError,
                              self._test_image_convert_invalid_vmdk)

    def test_image_convert_fails(self):
        convert = self._setup_image_convert_info_fail(disk_format='raw')
        with mock.patch.object(processutils, 'execute') as exc_mock:
            inspector = self.detect_file_format_mock.return_value
            inspector.__str__.return_value = 'raw'
            exc_mock.side_effect = [('{"format":"raw"}', ''),
                                    OSError('convert_fail')]
            self.assertRaises(OSError,
                              convert.execute, 'file:///test/path.raw')
            exc_mock.assert_has_calls(
                [mock.call('qemu-img', 'info',
                           '-f', 'raw',
                           '--output=json',
                           '/test/path.raw',
                           prlimit=async_utils.QEMU_IMG_PROC_LIMITS,
                           python_exec=convert.python,
                           log_errors=processutils.LOG_ALL_ERRORS),
                 mock.call('qemu-img', 'convert', '-f', 'raw', '-O', 'qcow2',
                           '/test/path.raw', '/test/path.raw.qcow2',
                           log_errors=processutils.LOG_ALL_ERRORS)])
        # Make sure we did not update the image
        self.img_repo.save.assert_not_called()

    def test_image_convert_reports_fail(self):
        convert = self._setup_image_convert_info_fail(disk_format='raw')
        with mock.patch.object(processutils, 'execute') as exc_mock:
            exc_mock.side_effect = [('{"format":"raw"}', ''),
                                    ('', 'some error')]
            inspector = self.detect_file_format_mock.return_value
            inspector.__str__.return_value = 'raw'
            self.assertRaises(RuntimeError,
                              convert.execute, 'file:///test/path.raw')
            exc_mock.assert_has_calls(
                [mock.call('qemu-img', 'info',
                           '-f', 'raw',
                           '--output=json',
                           '/test/path.raw',
                           prlimit=async_utils.QEMU_IMG_PROC_LIMITS,
                           python_exec=convert.python,
                           log_errors=processutils.LOG_ALL_ERRORS),
                 mock.call('qemu-img', 'convert', '-f', 'raw', '-O', 'qcow2',
                           '/test/path.raw', '/test/path.raw.qcow2',
                           log_errors=processutils.LOG_ALL_ERRORS)])
        # Make sure we did not update the image
        self.img_repo.save.assert_not_called()

    def test_image_convert_fails_source_format(self):
        convert = self._setup_image_convert_info_fail()
        with mock.patch.object(processutils, 'execute') as exc_mock:
            exc_mock.return_value = ('{}', '')
            inspector = self.detect_file_format_mock.return_value
            inspector.__str__.return_value = 'qcow2'
            exc = self.assertRaises(RuntimeError,
                                    convert.execute, 'file:///test/path.raw')
            self.assertIn('Image metadata disagrees about format', str(exc))
            exc_mock.assert_called_once_with(
                'qemu-img', 'info',
                '-f', 'qcow2',
                '--output=json',
                '/test/path.raw',
                prlimit=async_utils.QEMU_IMG_PROC_LIMITS,
                python_exec=convert.python,
                log_errors=processutils.LOG_ALL_ERRORS)
        # Make sure we did not update the image
        self.img_repo.save.assert_not_called()

    def test_image_convert_source_format_inspection_not_match(self):
        convert = self._setup_image_convert_info_fail(disk_format="raw")
        with mock.patch.object(processutils, 'execute') as exc_mock:
            exc_mock.return_value = (
                '{"format": "raw", "virtual-size": 123}', '')
            inspector = self.detect_file_format_mock.return_value
            inspector.__str__.return_value = 'qcow2'
            exc = self.assertRaises(RuntimeError,
                                    convert.execute, 'file:///test/path.raw')
            self.assertIn('Image format mismatch', str(exc))

    def test_image_convert_same_format_does_nothing(self):
        convert = self._setup_image_convert_info_fail()
        with mock.patch.object(processutils, 'execute') as exc_mock:
            exc_mock.return_value = (
                '{"format": "qcow2", "virtual-size": 123}', '')
            inspector = self.detect_file_format_mock.return_value
            inspector.__str__.return_value = 'qcow2'
            convert.execute('file:///test/path.qcow')
            # Make sure we only called qemu-img for inspection, not conversion
            exc_mock.assert_called_once_with(
                'qemu-img', 'info',
                '-f', 'qcow2',
                '--output=json',
                '/test/path.qcow',
                prlimit=async_utils.QEMU_IMG_PROC_LIMITS,
                python_exec=convert.python,
                log_errors=processutils.LOG_ALL_ERRORS)

        # Make sure we set the virtual_size before we exited
        image = self.img_repo.get.return_value
        self.assertEqual(123, image.virtual_size)

    @mock.patch.object(os, 'stat')
    @mock.patch.object(os, 'remove', new=mock.MagicMock())
    def _test_image_convert_inspect_target_format(
            self, mock_stat,
            out_format='raw', src_format='qcow2', dst_format='raw',
            dest_safe_fails=None):
        self.config(output_format=out_format,
                    group='image_conversion')
        convert = self._setup_image_convert_info_fail()
        mock_stat.return_value.st_size = 123
        with mock.patch.object(processutils, 'execute') as exc_mock:
            exc_mock.return_value = (
                '{"format": "%s", "virtual-size": 123}' % src_format, '')
            pre_inspector = mock.MagicMock()
            post_inspector = mock.MagicMock()
            self.detect_file_format_mock.side_effect = [
                pre_inspector, post_inspector]
            pre_inspector.__str__.return_value = pre_inspector.NAME = (
                src_format)
            post_inspector.__str__.return_value = post_inspector.NAME = (
                dst_format)
            if dest_safe_fails:
                post_inspector.safety_check.side_effect = (
                    format_inspector.SafetyCheckFailed(
                        {r: 'fail' for r in dest_safe_fails}))
            convert.execute('file:///test/path.%s' % src_format)

    def test_image_convert_ignores_configured_gpt_safety(self):
        # Conversion succeeds, finds the proper format, but fails safety check
        self.config(gpt_safety_checks_nonfatal=['mbr'],
                    group='image_format')

        # GPT that fails only mbr should be allowed to proceed
        self._test_image_convert_inspect_target_format(
            out_format='raw', dst_format='gpt', dest_safe_fails=['mbr'])

        # GPT that fails something not configured as non-fatal is rejected
        self.assertRaisesRegex(
            RuntimeError, 'disallowed configuration',
            self._test_image_convert_inspect_target_format,
            out_format='raw', dst_format='gpt', dest_safe_fails=['mbr',
                                                                 'pmbr'])

        # An non-GPT format that fails something configured as non-fatal for
        # GPT-only is rejected
        self.assertRaisesRegex(
            RuntimeError, 'disallowed configuration',
            self._test_image_convert_inspect_target_format,
            out_format='vmdk', dst_format='vmdk', dest_safe_fails=['mbr'])

    def test_image_convert_fails_target_format_raw(self):
        # Uploaded a qcow with a vmdk inside it. Conversion expects a
        # raw, but finds the wrong type and should abort
        self.assertRaisesRegex(
            RuntimeError, 'image in unexpected format',
            self._test_image_convert_inspect_target_format,
            dst_format='vmdk')

    def test_image_convert_fails_target_format_unsafe(self):
        # Conversion succeeds, finds the proper format, but fails safety check
        self.assertRaisesRegex(
            RuntimeError, 'disallowed configuration',
            self._test_image_convert_inspect_target_format,
            out_format='vmdk', dst_format='vmdk', dest_safe_fails=['header'])

    def _set_image_conversion(self, mock_os_remove, stores=[]):
        mock_os_remove.return_value = None
        wrapper = mock.MagicMock()
        image_convert = image_conversion._ConvertImage(self.context,
                                                       self.task.task_id,
                                                       self.task_type,
                                                       wrapper,
                                                       stores)
        action = wrapper.__enter__.return_value
        self.task_repo.get.return_value = self.task
        return action, image_convert

    @mock.patch.object(os, 'remove')
    def test_image_convert_revert_success_multiple_stores(
            self, mock_os_remove):
        action, image_convert = self._set_image_conversion(
            mock_os_remove, stores=self.stores)

        with mock.patch.object(processutils, 'execute') as exc_mock:
            exc_mock.return_value = ("", None)
            with mock.patch.object(os.path, 'exists') as os_exists_mock:
                os_exists_mock.return_value = True
                image_convert.revert(result=mock.MagicMock())
                self.assertEqual(1, mock_os_remove.call_count)
                action.set_image_attribute.assert_called_once_with(
                    status='queued')
                action.remove_importing_stores.assert_called_once_with(
                    self.stores)
                action.add_failed_stores.assert_called_once_with(
                    self.stores)

    @mock.patch.object(os, 'remove')
    def test_image_convert_revert_success_single_store(
            self, mock_os_remove):
        action, image_convert = self._set_image_conversion(mock_os_remove)

        with mock.patch.object(processutils, 'execute') as exc_mock:
            exc_mock.return_value = ("", None)
            with mock.patch.object(os.path, 'exists') as os_exists_mock:
                os_exists_mock.return_value = True
                image_convert.revert(result=mock.MagicMock())
                self.assertEqual(1, mock_os_remove.call_count)
                self.assertEqual(0, action.remove_importing_stores.call_count)
                self.assertEqual(0, action.add_failed_store.call_count)
                action.set_image_attribute.assert_called_once_with(
                    status='queued')

    @mock.patch.object(os, 'remove')
    def test_image_convert_revert_success_src_file_exists(
            self, mock_os_remove):
        action, image_convert = self._set_image_conversion(
            mock_os_remove, stores=self.stores)
        image_convert.src_path = mock.MagicMock()

        with mock.patch.object(processutils, 'execute') as exc_mock:
            exc_mock.return_value = ("", None)
            with mock.patch.object(os.path, 'exists') as os_exists_mock:
                os_exists_mock.return_value = True
                image_convert.revert(result=mock.MagicMock())
                action.set_image_attribute.assert_called_once_with(
                    status='queued')
                action.remove_importing_stores.assert_called_once_with(
                    self.stores)
                action.add_failed_stores.assert_called_once_with(
                    self.stores)
                self.assertEqual(2, mock_os_remove.call_count)

    def test_image_convert_interpreter_configured(self):
        # By default, wsgi.python_interpreter is None; if it is
        # overridden, we should take the interpreter from config.
        fake_interpreter = '/usr/bin/python2.7'
        self.config(python_interpreter=fake_interpreter,
                    group='wsgi')
        convert = image_conversion._ConvertImage(self.context,
                                                 self.task.task_id,
                                                 self.task_type,
                                                 self.wrapper,
                                                 self.stores)
        self.assertEqual(fake_interpreter, convert.python)



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_log.fixture import logging_error as log_fixture
import testtools
import webob

import glance.api.common
from glance.common import exception
from glance.tests.unit import fixtures as glance_fixtures


class SimpleIterator(object):
    def __init__(self, file_object, chunk_size):
        self.file_object = file_object
        self.chunk_size = chunk_size

    def __iter__(self):
        def read_chunk():
            return self.fobj.read(self.chunk_size)

        chunk = read_chunk()
        while chunk:
            yield chunk
            chunk = read_chunk()
        else:
            return


class TestSizeCheckedIter(testtools.TestCase):

    def setUp(self):
        super().setUp()

        # Limit the amount of DeprecationWarning messages in the unit test logs
        self.useFixture(glance_fixtures.WarningsFixture())

        # Make sure logging output is limited but still test debug formatting
        self.useFixture(log_fixture.get_logging_handle_error_fixture())
        self.useFixture(glance_fixtures.StandardLogging())

    def _get_image_metadata(self):
        return {'id': 'e31cb99c-fe89-49fb-9cc5-f5104fffa636'}

    def _get_webob_response(self):
        request = webob.Request.blank('/')
        response = webob.Response()
        response.request = request
        return response

    def test_uniform_chunk_size(self):
        resp = self._get_webob_response()
        meta = self._get_image_metadata()
        checked_image = glance.api.common.size_checked_iter(
            resp, meta, 4, ['AB', 'CD'], None)

        self.assertEqual('AB', next(checked_image))
        self.assertEqual('CD', next(checked_image))
        self.assertRaises(StopIteration, next, checked_image)

    def test_small_last_chunk(self):
        resp = self._get_webob_response()
        meta = self._get_image_metadata()
        checked_image = glance.api.common.size_checked_iter(
            resp, meta, 3, ['AB', 'C'], None)

        self.assertEqual('AB', next(checked_image))
        self.assertEqual('C', next(checked_image))
        self.assertRaises(StopIteration, next, checked_image)

    def test_variable_chunk_size(self):
        resp = self._get_webob_response()
        meta = self._get_image_metadata()
        checked_image = glance.api.common.size_checked_iter(
            resp, meta, 6, ['AB', '', 'CDE', 'F'], None)

        self.assertEqual('AB', next(checked_image))
        self.assertEqual('', next(checked_image))
        self.assertEqual('CDE', next(checked_image))
        self.assertEqual('F', next(checked_image))
        self.assertRaises(StopIteration, next, checked_image)

    def test_too_many_chunks(self):
        """An image should streamed regardless of expected_size"""
        resp = self._get_webob_response()
        meta = self._get_image_metadata()
        checked_image = glance.api.common.size_checked_iter(
            resp, meta, 4, ['AB', 'CD', 'EF'], None)

        self.assertEqual('AB', next(checked_image))
        self.assertEqual('CD', next(checked_image))
        self.assertEqual('EF', next(checked_image))
        self.assertRaises(exception.GlanceException, next, checked_image)

    def test_too_few_chunks(self):
        resp = self._get_webob_response()
        meta = self._get_image_metadata()
        checked_image = glance.api.common.size_checked_iter(resp, meta, 6,
                                                            ['AB', 'CD'],
                                                            None)

        self.assertEqual('AB', next(checked_image))
        self.assertEqual('CD', next(checked_image))
        self.assertRaises(exception.GlanceException, next, checked_image)

    def test_too_much_data(self):
        resp = self._get_webob_response()
        meta = self._get_image_metadata()
        checked_image = glance.api.common.size_checked_iter(resp, meta, 3,
                                                            ['AB', 'CD'],
                                                            None)

        self.assertEqual('AB', next(checked_image))
        self.assertEqual('CD', next(checked_image))
        self.assertRaises(exception.GlanceException, next, checked_image)

    def test_too_little_data(self):
        resp = self._get_webob_response()
        meta = self._get_image_metadata()
        checked_image = glance.api.common.size_checked_iter(resp, meta, 6,
                                                            ['AB', 'CD', 'E'],
                                                            None)

        self.assertEqual('AB', next(checked_image))
        self.assertEqual('CD', next(checked_image))
        self.assertEqual('E', next(checked_image))
        self.assertRaises(exception.GlanceException, next, checked_image)


class TestThreadPool(testtools.TestCase):
    def setUp(self):
        super().setUp()

        # Limit the amount of DeprecationWarning messages in the unit test logs
        self.useFixture(glance_fixtures.WarningsFixture())

        # Make sure logging output is limited but still test debug formatting
        self.useFixture(log_fixture.get_logging_handle_error_fixture())
        self.useFixture(glance_fixtures.StandardLogging())

    @mock.patch('glance.async_.get_threadpool_model')
    def test_get_thread_pool(self, mock_gtm):
        get_thread_pool = glance.api.common.get_thread_pool

        pool1 = get_thread_pool('pool1', size=123)
        get_thread_pool('pool2', size=456)
        pool1a = get_thread_pool('pool1')

        # Two calls for the same pool should return the exact same thing
        self.assertEqual(pool1, pool1a)

        # Only two calls to get new threadpools should have been made
        mock_gtm.return_value.assert_has_calls(
            [mock.call(123), mock.call(456)])

    @mock.patch('glance.async_.get_threadpool_model')
    def test_get_thread_pool_log(self, mock_gtm):
        with mock.patch.object(glance.api.common, 'LOG') as mock_log:
            glance.api.common.get_thread_pool('test-pool')
            mock_log.debug.assert_called_once_with(
                'Initializing named threadpool %r', 'test-pool')



# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from glance.common import wsgi


def init(mapper):
    reject_resource = wsgi.Resource(wsgi.RejectMethodController())
    mapper.connect("/v1", controller=reject_resource,
                   action="reject")


class API(wsgi.Router):
    """WSGI entry point for satisfy grenade."""

    def __init__(self, mapper):
        mapper = mapper or wsgi.APIMapper()

        init(mapper)

        super(API, self).__init__(mapper)



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import io

from cryptography import exceptions as crypto_exception
from cursive import exception as cursive_exception
from cursive import signature_utils
import glance_store
from oslo_config import cfg
import struct
from unittest import mock

from glance.common import exception
import glance.location
from glance.tests.unit import base as unit_test_base
from glance.tests.unit import utils as unit_test_utils
from glance.tests import utils


BASE_URI = 'http://storeurl.com/container'
UUID1 = 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d'
UUID2 = '971ec09a-8067-4bc8-a91f-ae3557f1c4c7'
USER1 = '54492ba0-f4df-4e4e-be62-27f4d76b29cf'
TENANT1 = '6838eb7b-6ded-434a-882c-b344c77fe8df'
TENANT2 = '2c014f32-55eb-467d-8fcb-4bd706012f81'
TENANT3 = '228c6da5-29cd-4d67-9457-ed632e083fc0'
CONF = cfg.CONF


class ImageRepoStub(object):
    def add(self, image):
        return image

    def save(self, image, from_state=None):
        return image


class ImageStub(object):
    def __init__(self, image_id, status=None, locations=None,
                 visibility=None, extra_properties=None, virtual_size=0):
        self.image_id = image_id
        self.status = status
        self.locations = locations or []
        self.visibility = visibility
        self.size = None
        self.extra_properties = extra_properties or {}
        self.os_hash_algo = None
        self.os_hash_value = None
        self.checksum = None
        self.disk_format = 'raw'
        self.container_format = 'bare'
        self.virtual_size = virtual_size

    def delete(self):
        self.status = 'deleted'

    def get_member_repo(self):
        return FakeMemberRepo(self, [TENANT1, TENANT2])


class ImageFactoryStub(object):
    def new_image(self, image_id=None, name=None, visibility='private',
                  min_disk=0, min_ram=0, protected=False, owner=None,
                  disk_format=None, container_format=None,
                  extra_properties=None, tags=None, **other_args):
        return ImageStub(image_id, visibility=visibility,
                         extra_properties=extra_properties, **other_args)


class FakeMemberRepo(object):
    def __init__(self, image, tenants=None):
        self.image = image
        self.factory = glance.domain.ImageMemberFactory()
        self.tenants = tenants or []

    def list(self, *args, **kwargs):
        return [self.factory.new_image_member(self.image, tenant)
                for tenant in self.tenants]

    def add(self, member):
        self.tenants.append(member.member_id)

    def remove(self, member):
        self.tenants.remove(member.member_id)


class TestStoreMultiBackends(utils.BaseTestCase):
    def setUp(self):
        self.store_api = unit_test_utils.FakeStoreAPI()
        self.store_utils = unit_test_utils.FakeStoreUtils(self.store_api)
        self.enabled_backends = {
            "ceph1": "rbd",
            "ceph2": "rbd"
        }
        super(TestStoreMultiBackends, self).setUp()
        self.config(enabled_backends=self.enabled_backends)

    @mock.patch("glance.location.signature_utils.get_verifier")
    def test_set_data_calls_upload_to_store(self, msig):
        context = glance.context.RequestContext(user=USER1)
        extra_properties = {
            'img_signature_certificate_uuid': 'UUID',
            'img_signature_hash_method': 'METHOD',
            'img_signature_key_type': 'TYPE',
            'img_signature': 'VALID'
        }
        image_stub = ImageStub(UUID2, status='queued', locations=[],
                               extra_properties=extra_properties,
                               virtual_size=4)
        image_stub.disk_format = 'raw'
        image = glance.location.ImageProxy(image_stub, context,
                                           self.store_api, self.store_utils)
        found_data = []

        def consumer(data, _v, _b, _s):
            found_data.append(data.read(4))

        with mock.patch.object(image, "_upload_to_store") as mloc:
            mloc.side_effect = consumer
            src_data = io.BytesIO(b'YYYY')
            image.set_data(src_data, 4, backend='ceph1')
            msig.assert_called_once_with(context=context,
                                         img_signature_certificate_uuid='UUID',
                                         img_signature_hash_method='METHOD',
                                         img_signature='VALID',
                                         img_signature_key_type='TYPE')
            mloc.assert_called_once_with(mock.ANY, msig.return_value, 'ceph1',
                                         4)
            self.assertEqual([b'YYYY'], found_data)

        self.assertEqual('active', image.status)

    def test_image_set_data(self):
        store_api = mock.MagicMock()
        store_api.add_with_multihash.return_value = (
            "rbd://ceph1", 4, "Z", "MH", {"backend": "ceph1"})
        context = glance.context.RequestContext(user=USER1)
        image_stub = ImageStub(UUID2, status='queued', locations=[])
        image = glance.location.ImageProxy(image_stub, context,
                                           store_api, self.store_utils)
        image.set_data('YYYY', 4, backend='ceph1')
        self.assertEqual(4, image.size)

        # NOTE(markwash): FakeStore returns image_id for location
        self.assertEqual("rbd://ceph1", image.locations[0]['url'])
        self.assertEqual({"backend": "ceph1"}, image.locations[0]['metadata'])
        self.assertEqual('Z', image.checksum)
        self.assertEqual('active', image.status)

    @mock.patch('glance.location.LOG')
    def test_image_set_data_valid_signature(self, mock_log):
        store_api = mock.MagicMock()
        store_api.add_with_multihash.return_value = (
            "rbd://ceph1", 4, "Z", "MH", {"backend": "ceph1"})
        context = glance.context.RequestContext(user=USER1)
        extra_properties = {
            'img_signature_certificate_uuid': 'UUID',
            'img_signature_hash_method': 'METHOD',
            'img_signature_key_type': 'TYPE',
            'img_signature': 'VALID'
        }
        image_stub = ImageStub(UUID2, status='queued',
                               extra_properties=extra_properties)
        self.mock_object(signature_utils, 'get_verifier',
                         unit_test_utils.fake_get_verifier)
        image = glance.location.ImageProxy(image_stub, context,
                                           store_api, self.store_utils)
        image.set_data('YYYY', 4, backend='ceph1')
        self.assertEqual('active', image.status)
        call = mock.call('Successfully verified signature for image %s',
                         UUID2)
        mock_log.info.assert_has_calls([call])

    @mock.patch("glance.location.signature_utils.get_verifier")
    def test_image_set_data_invalid_signature(self, msig):
        msig.return_value.verify.side_effect = \
            crypto_exception.InvalidSignature
        store_api = mock.MagicMock()
        store_api.add_with_multihash.return_value = (
            "rbd://ceph1", 4, "Z", "MH", {"backend": "ceph1"})
        context = glance.context.RequestContext(user=USER1)
        extra_properties = {
            'img_signature_certificate_uuid': 'UUID',
            'img_signature_hash_method': 'METHOD',
            'img_signature_key_type': 'TYPE',
            'img_signature': 'INVALID'
        }
        image_stub = ImageStub(UUID2, status='queued',
                               extra_properties=extra_properties)
        image = glance.location.ImageProxy(image_stub, context,
                                           store_api, self.store_utils)
        self.assertRaises(cursive_exception.SignatureVerificationError,
                          image.set_data, 'YYYY', 4, backend='ceph1')


class TestStoreImage(utils.BaseTestCase):
    def setUp(self):
        locations = [{'url': '%s/%s' % (BASE_URI, UUID1),
                      'metadata': {}, 'status': 'active'}]
        self.image_stub = ImageStub(UUID1, 'active', locations)
        self.store_api = unit_test_utils.FakeStoreAPI()
        self.store_utils = unit_test_utils.FakeStoreUtils(self.store_api)
        super(TestStoreImage, self).setUp()

    def test_image_delete(self):
        image = glance.location.ImageProxy(self.image_stub, {},
                                           self.store_api, self.store_utils)
        location = image.locations[0]
        self.assertEqual('active', image.status)
        self.store_api.get_from_backend(location['url'], context={})
        image.delete()
        self.assertEqual('deleted', image.status)
        self.assertRaises(glance_store.NotFound,
                          self.store_api.get_from_backend, location['url'], {})

    def test_image_get_data(self):
        image = glance.location.ImageProxy(self.image_stub, {},
                                           self.store_api, self.store_utils)
        self.assertEqual('XXX', image.get_data())

    def test_image_get_data_from_second_location(self):
        def fake_get_from_backend(self, location, offset=0,
                                  chunk_size=None, context=None):
            if UUID1 in location:
                raise Exception('not allow download from %s' % location)
            else:
                return self.data[location]

        image1 = glance.location.ImageProxy(self.image_stub, {},
                                            self.store_api, self.store_utils)
        self.assertEqual('XXX', image1.get_data())
        # Multiple location support
        context = glance.context.RequestContext(user=USER1)
        (image2, image_stub2) = self._add_image(context, UUID2, 'ZZZ', 3)
        location_data = image2.locations[0]

        with mock.patch("glance.location.store") as mock_store:
            mock_store.get_size_from_uri_and_backend.return_value = 3
            image1.locations.append(location_data)

        self.assertEqual(2, len(image1.locations))
        self.assertEqual(UUID2, location_data['url'])

        self.mock_object(unit_test_utils.FakeStoreAPI, 'get_from_backend',
                         fake_get_from_backend)
        # This time, image1.get_data() returns the data wrapped in a
        # LimitingReader|CooperativeReader|InfoWrapper pipeline, so
        # peeking under the hood of those objects to get at the
        # underlying string.
        self.assertEqual('ZZZ', image1.get_data().data.fd._source)

        image1.locations.pop(0)
        self.assertEqual(1, len(image1.locations))
        image2.delete()

    def test_image_set_data(self):
        context = glance.context.RequestContext(user=USER1)
        image_stub = ImageStub(UUID2, status='queued', locations=[])
        # We are going to pass an iterable data source, so use the
        # FakeStoreAPIReader that actually reads from that data
        store_api = unit_test_utils.FakeStoreAPIReader()
        image = glance.location.ImageProxy(image_stub, context,
                                           store_api, self.store_utils)
        image.set_data(iter(['YYYY']), 4)
        self.assertEqual(4, image.size)
        # NOTE(markwash): FakeStore returns image_id for location
        self.assertEqual(UUID2, image.locations[0]['url'])
        self.assertEqual('Z', image.checksum)
        self.assertEqual('active', image.status)
        self.assertEqual(4, image.virtual_size)

    def test_image_set_data_inspector_no_match(self):
        context = glance.context.RequestContext(user=USER1)
        image_stub = ImageStub(UUID2, status='queued', locations=[])
        image_stub.disk_format = 'qcow2'
        # We are going to pass an iterable data source, so use the
        # FakeStoreAPIReader that actually reads from that data
        store_api = unit_test_utils.FakeStoreAPIReader()
        image = glance.location.ImageProxy(image_stub, context,
                                           store_api, self.store_utils)
        if CONF.image_format.require_image_format_match:
            self.assertRaises(exception.InvalidImageData,
                              image.set_data, iter(['YYYY']), 4)
        else:
            image.set_data(iter(['YYYY']), 4)
            self.assertEqual(4, image.size)
            # NOTE(markwash): FakeStore returns image_id for location
            self.assertEqual(UUID2, image.locations[0]['url'])
            self.assertEqual('Z', image.checksum)
            self.assertEqual('active', image.status)
            self.assertEqual(0, image.virtual_size)

    def test_image_set_data_inspector_unsupported_format(self):
        context = glance.context.RequestContext(user=USER1)
        image_stub = ImageStub(UUID2, status='queued', locations=[])
        image_stub.disk_format = 'ami'
        # We are going to pass an iterable data source, so use the
        # FakeStoreAPIReader that actually reads from that data
        store_api = unit_test_utils.FakeStoreAPIReader()
        image = glance.location.ImageProxy(image_stub, context,
                                           store_api, self.store_utils)
        image.set_data(iter(['YYYY']), 4)
        self.assertEqual(4, image.size)
        # NOTE(markwash): FakeStore returns image_id for location
        self.assertEqual(UUID2, image.locations[0]['url'])
        self.assertEqual('Z', image.checksum)
        self.assertEqual('active', image.status)
        self.assertEqual(0, image.virtual_size)

    def _get_stub_gpt(self):
        """Returns a valid, safe GPT-based image"""
        data = bytearray(b'\x00' * 1024)
        # The last two bytes of the first sector is the little-endian signature
        # value 0xAA55
        data[510:512] = b'\x55\xAA'

        # This is one EFI Protective MBR partition in the first PTE slot,
        # which is 16 bytes starting at offset 446.
        data[446:446 + 16] = struct.pack('<BBBBBBBBII',
                                         0x00,  # boot
                                         0x00,  # start C
                                         0x02,  # start H
                                         0x00,  # start S
                                         0xEE,  # OS type
                                         0x00,  # end C
                                         0x00,  # end H
                                         0x00,  # end S
                                         0x01,  # start LBA
                                         0x00,  # size LBA
                                         )
        return data

    def test_image_set_data_inspector_gpt_as_raw_compat(self):
        context = glance.context.RequestContext(user=USER1)
        image_stub = ImageStub(UUID2, status='queued', locations=[])
        image_stub.disk_format = 'raw'
        # We are going to pass an iterable data source, so use the
        # FakeStoreAPIReader that actually reads from that data

        store_api = unit_test_utils.FakeStoreAPIReader(max_size=2048)
        image = glance.location.ImageProxy(image_stub, context,
                                           store_api, self.store_utils)
        data = self._get_stub_gpt()
        image.set_data(iter((data,)), 1024)
        self.assertEqual(1024, image.size)
        # NOTE(markwash): FakeStore returns image_id for location
        self.assertEqual(UUID2, image.locations[0]['url'])
        self.assertEqual('Z', image.checksum)
        self.assertEqual('active', image.status)
        self.assertEqual(1024, image.virtual_size)

    def test_image_set_data_inspector_format_unsupported_hiding(self):
        context = glance.context.RequestContext(user=USER1)
        image_stub = ImageStub(UUID2, status='queued', locations=[])
        image_stub.disk_format = 'ami'
        # We are going to pass an iterable data source, so use the
        # FakeStoreAPIReader that actually reads from that data
        data = self._get_stub_gpt()
        store_api = unit_test_utils.FakeStoreAPIReader(max_size=2048)
        image = glance.location.ImageProxy(image_stub, context,
                                           store_api, self.store_utils)
        # Image was claimed to be AMI, which we cannot inspect, but we found
        # a known format inside, which is invalid
        self.assertRaises(exception.InvalidImageData,
                          image.set_data, iter((data,)), 1024)

    def test_image_set_data_inspector_format_fails_safety_check(self):
        self.config(gpt_safety_checks_nonfatal=[], group='image_format')
        context = glance.context.RequestContext(user=USER1)
        image_stub = ImageStub(UUID2, status='queued', locations=[])
        image_stub.disk_format = 'gpt'
        # We are going to pass an iterable data source, so use the
        # FakeStoreAPIReader that actually reads from that data
        data = self._get_stub_gpt()
        # Zero out the partition table to make this unsafe
        data[446:446 + 16] = b'\x00' * 16
        store_api = unit_test_utils.FakeStoreAPIReader(max_size=2048)
        image = glance.location.ImageProxy(image_stub, context,
                                           store_api, self.store_utils)
        # Image was claimed to be AMI, which we cannot inspect, but we found
        # a known format inside, which is invalid
        self.assertRaisesRegex(exception.InvalidImageData,
                               'safety checks',
                               image.set_data, iter((data,)), 1024)

    def test_image_set_data_gpt_safety_ignored(self):
        self.config(gpt_safety_checks_nonfatal=['mbr'], group='image_format')
        context = glance.context.RequestContext(user=USER1)
        image_stub = ImageStub(UUID2, status='queued', locations=[])
        image_stub.disk_format = 'gpt'
        # We are going to pass an iterable data source, so use the
        # FakeStoreAPIReader that actually reads from that data
        data = self._get_stub_gpt()
        # Zero out the partition table to make this unsafe
        data[446:446 + 16] = b'\x00' * 16
        store_api = unit_test_utils.FakeStoreAPIReader(max_size=2048)
        image = glance.location.ImageProxy(image_stub, context,
                                           store_api, self.store_utils)
        image.set_data(iter((data,)), 1024)

    def test_image_set_data_inspector_no_match_disabled(self):
        self.config(require_image_format_match=False, group='image_format')
        self.test_image_set_data_inspector_no_match()

    @mock.patch('oslo_utils.imageutils.format_inspector.QcowInspector.'
                'virtual_size',
                new_callable=mock.PropertyMock)
    @mock.patch('oslo_utils.imageutils.format_inspector.QcowInspector.'
                'format_match',
                new_callable=mock.PropertyMock)
    def test_image_set_data_inspector_virtual_size_failure(self, mock_fm,
                                                           mock_vs):
        # Force our format to match
        mock_fm.return_value = True

        # Make virtual_size fail in some unexpected way
        mock_vs.side_effect = ValueError('some error')

        context = glance.context.RequestContext(user=USER1)
        image_stub = ImageStub(UUID2, status='queued', locations=[])
        image_stub.disk_format = 'qcow2'
        # We are going to pass an iterable data source, so use the
        # FakeStoreAPIReader that actually reads from that data
        store_api = unit_test_utils.FakeStoreAPIReader()
        image = glance.location.ImageProxy(image_stub, context,
                                           store_api, self.store_utils)

        if CONF.image_format.require_image_format_match:
            self.assertRaises(exception.InvalidImageData,
                              image.set_data, iter(['YYYY']), 4)
        else:
            # Make sure set_data proceeds even though the format clearly
            # does not match
            image.set_data(iter(['YYYY']), 4)
            self.assertEqual(4, image.size)
            # NOTE(markwash): FakeStore returns image_id for location
            self.assertEqual(UUID2, image.locations[0]['url'])
            self.assertEqual('Z', image.checksum)
            self.assertEqual('active', image.status)
            self.assertEqual(0, image.virtual_size)

    def test_image_set_data_inspector_virtual_size_failure_disabled(self):
        self.config(require_image_format_match=False, group='image_format')
        self.test_image_set_data_inspector_virtual_size_failure()

    @mock.patch('oslo_utils.imageutils.format_inspector.InspectWrapper')
    def test_image_set_data_inspector_not_needed(self, mock_inspect):
        context = glance.context.RequestContext(user=USER1)
        image_stub = ImageStub(UUID2, status='queued', locations=[])
        image_stub.virtual_size = 123
        image_stub.container_format = 'gzip'
        image_stub.disk_format = 'qcow2'
        # We are going to pass an iterable data source, so use the
        # FakeStoreAPIReader that actually reads from that data
        store_api = unit_test_utils.FakeStoreAPIReader()
        image = glance.location.ImageProxy(image_stub, context,
                                           store_api, self.store_utils)
        image.set_data(iter(['YYYY']), 4)
        self.assertEqual(4, image.size)
        # NOTE(markwash): FakeStore returns image_id for location
        self.assertEqual(UUID2, image.locations[0]['url'])
        self.assertEqual('Z', image.checksum)
        self.assertEqual('active', image.status)
        self.assertEqual(123, image.virtual_size)
        # If the image already had virtual_size set (i.e. we're setting
        # a new location), we should not re-calculate the value.
        mock_inspect.assert_not_called()

    def test_image_set_data_location_metadata(self):
        context = glance.context.RequestContext(user=USER1)
        image_stub = ImageStub(UUID2, status='queued', locations=[])
        loc_meta = {'key': 'value5032'}
        store_api = unit_test_utils.FakeStoreAPI(store_metadata=loc_meta)
        store_utils = unit_test_utils.FakeStoreUtils(store_api)
        image = glance.location.ImageProxy(image_stub, context,
                                           store_api, store_utils)
        image.set_data('YYYY', 4)
        self.assertEqual(4, image.size)
        location_data = image.locations[0]
        self.assertEqual(UUID2, location_data['url'])
        self.assertEqual(loc_meta, location_data['metadata'])
        self.assertEqual('Z', image.checksum)
        self.assertEqual('active', image.status)
        image.delete()
        self.assertEqual(image.status, 'deleted')
        self.assertRaises(glance_store.NotFound,
                          self.store_api.get_from_backend,
                          image.locations[0]['url'], {})

    def test_image_set_data_unknown_size(self):
        context = glance.context.RequestContext(user=USER1)
        image_stub = ImageStub(UUID2, status='queued', locations=[])
        image_stub.disk_format = 'iso'
        image = glance.location.ImageProxy(image_stub, context,
                                           self.store_api, self.store_utils)

        if CONF.image_format.require_image_format_match:
            self.assertRaises(exception.InvalidImageData,
                              image.set_data, 'YYYY', None)
        else:
            image.set_data('YYYY', None)
            self.assertEqual(4, image.size)
            # NOTE(markwash): FakeStore returns image_id for location
            self.assertEqual(UUID2, image.locations[0]['url'])
            self.assertEqual('Z', image.checksum)
            self.assertEqual('active', image.status)
            image.delete()
            self.assertEqual(image.status, 'deleted')
            self.assertRaises(glance_store.NotFound,
                              self.store_api.get_from_backend,
                              image.locations[0]['url'], context={})

    def test_image_set_data_unknown_size_disabled(self):
        self.config(require_image_format_match=False, group='image_format')
        self.test_image_set_data_unknown_size()

    @mock.patch('glance.location.LOG')
    def test_image_set_data_valid_signature(self, mock_log):
        context = glance.context.RequestContext(user=USER1)
        extra_properties = {
            'img_signature_certificate_uuid': 'UUID',
            'img_signature_hash_method': 'METHOD',
            'img_signature_key_type': 'TYPE',
            'img_signature': 'VALID'
        }
        image_stub = ImageStub(UUID2, status='queued',
                               extra_properties=extra_properties)
        self.mock_object(signature_utils, 'get_verifier',
                         unit_test_utils.fake_get_verifier)
        image = glance.location.ImageProxy(image_stub, context,
                                           self.store_api, self.store_utils)
        image.set_data('YYYY', 4)
        self.assertEqual('active', image.status)
        mock_log.info.assert_any_call(
            'Successfully verified signature for image %s',
            UUID2)

    def test_image_set_data_invalid_signature(self):
        context = glance.context.RequestContext(user=USER1)
        extra_properties = {
            'img_signature_certificate_uuid': 'UUID',
            'img_signature_hash_method': 'METHOD',
            'img_signature_key_type': 'TYPE',
            'img_signature': 'INVALID'
        }
        image_stub = ImageStub(UUID2, status='queued',
                               extra_properties=extra_properties)
        self.mock_object(signature_utils, 'get_verifier',
                         unit_test_utils.fake_get_verifier)
        image = glance.location.ImageProxy(image_stub, context,
                                           self.store_api, self.store_utils)
        with mock.patch.object(self.store_api,
                               'delete_from_backend') as mock_delete:
            self.assertRaises(cursive_exception.SignatureVerificationError,
                              image.set_data,
                              'YYYY', 4)
            mock_delete.assert_called()

    def test_image_set_data_invalid_signature_missing_metadata(self):
        context = glance.context.RequestContext(user=USER1)
        extra_properties = {
            'img_signature_hash_method': 'METHOD',
            'img_signature_key_type': 'TYPE',
            'img_signature': 'INVALID'
        }
        image_stub = ImageStub(UUID2, status='queued',
                               extra_properties=extra_properties)
        self.mock_object(signature_utils, 'get_verifier',
                         unit_test_utils.fake_get_verifier)
        image = glance.location.ImageProxy(image_stub, context,
                                           self.store_api, self.store_utils)
        image.set_data('YYYY', 4)
        self.assertEqual(UUID2, image.locations[0]['url'])
        self.assertEqual('Z', image.checksum)
        # Image is still active, since invalid signature was ignored
        self.assertEqual('active', image.status)

    def _add_image(self, context, image_id, data, len):
        image_stub = ImageStub(image_id, status='queued', locations=[])
        image = glance.location.ImageProxy(image_stub, context,
                                           self.store_api, self.store_utils)
        image.set_data(data, len)
        self.assertEqual(len, image.size)
        # NOTE(markwash): FakeStore returns image_id for location
        location = {'url': image_id, 'metadata': {}, 'status': 'active'}
        self.assertEqual([location], image.locations)
        self.assertEqual([location], image_stub.locations)
        self.assertEqual('active', image.status)
        return (image, image_stub)

    def test_image_change_append_invalid_location_uri(self):
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)

        location_bad = {'url': 'unknown://location', 'metadata': {}}
        self.assertRaises(exception.BadStoreUri,
                          image1.locations.append, location_bad)

        image1.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())

    def test_image_change_append_invalid_location_metatdata(self):
        UUID3 = 'a8a61ec4-d7a3-11e2-8c28-000c29c27581'
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)
        (image2, image_stub2) = self._add_image(context, UUID3, 'YYYY', 4)

        # Using only one test rule here is enough to make sure
        # 'store.check_location_metadata()' can be triggered
        # in Location proxy layer. Complete test rule for
        # 'store.check_location_metadata()' testing please
        # check below cases within 'TestStoreMetaDataChecker'.
        location_bad = {'url': UUID3, 'metadata': b"a invalid metadata"}

        self.assertRaises(glance_store.BackendException,
                          image1.locations.append, location_bad)

        image1.delete()
        image2.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())
        self.assertNotIn(UUID3, self.store_api.data.keys())

    def test_image_change_append_locations(self):
        UUID3 = 'a8a61ec4-d7a3-11e2-8c28-000c29c27581'
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)
        (image2, image_stub2) = self._add_image(context, UUID3, 'YYYY', 4)

        location2 = {'url': UUID2, 'metadata': {}, 'status': 'active'}
        location3 = {'url': UUID3, 'metadata': {}, 'status': 'active'}

        image1.locations.append(location3)

        self.assertEqual([location2, location3], image_stub1.locations)
        self.assertEqual([location2, location3], image1.locations)

        image1.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())
        self.assertNotIn(UUID3, self.store_api.data.keys())

        image2.delete()

    def test_image_change_pop_location(self):
        UUID3 = 'a8a61ec4-d7a3-11e2-8c28-000c29c27581'
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)
        (image2, image_stub2) = self._add_image(context, UUID3, 'YYYY', 4)

        location2 = {'url': UUID2, 'metadata': {}, 'status': 'active'}
        location3 = {'url': UUID3, 'metadata': {}, 'status': 'active'}

        image1.locations.append(location3)

        self.assertEqual([location2, location3], image_stub1.locations)
        self.assertEqual([location2, location3], image1.locations)

        image1.locations.pop()

        self.assertEqual([location2], image_stub1.locations)
        self.assertEqual([location2], image1.locations)

        image1.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())
        self.assertNotIn(UUID3, self.store_api.data.keys())

        image2.delete()

    def test_image_change_extend_invalid_locations_uri(self):
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)

        location_bad = {'url': 'unknown://location', 'metadata': {}}

        self.assertRaises(exception.BadStoreUri,
                          image1.locations.extend, [location_bad])

        image1.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())

    def test_image_change_extend_invalid_locations_metadata(self):
        UUID3 = 'a8a61ec4-d7a3-11e2-8c28-000c29c27581'
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)
        (image2, image_stub2) = self._add_image(context, UUID3, 'YYYY', 4)

        location_bad = {'url': UUID3, 'metadata': b"a invalid metadata"}

        self.assertRaises(glance_store.BackendException,
                          image1.locations.extend, [location_bad])

        image1.delete()
        image2.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())
        self.assertNotIn(UUID3, self.store_api.data.keys())

    def test_image_change_extend_locations(self):
        UUID3 = 'a8a61ec4-d7a3-11e2-8c28-000c29c27581'
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)
        (image2, image_stub2) = self._add_image(context, UUID3, 'YYYY', 4)

        location2 = {'url': UUID2, 'metadata': {}, 'status': 'active'}
        location3 = {'url': UUID3, 'metadata': {}, 'status': 'active'}

        image1.locations.extend([location3])

        self.assertEqual([location2, location3], image_stub1.locations)
        self.assertEqual([location2, location3], image1.locations)

        image1.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())
        self.assertNotIn(UUID3, self.store_api.data.keys())

        image2.delete()

    def test_image_change_remove_location(self):
        UUID3 = 'a8a61ec4-d7a3-11e2-8c28-000c29c27581'
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)
        (image2, image_stub2) = self._add_image(context, UUID3, 'YYYY', 4)

        location2 = {'url': UUID2, 'metadata': {}, 'status': 'active'}
        location3 = {'url': UUID3, 'metadata': {}, 'status': 'active'}
        location_bad = {'url': 'unknown://location', 'metadata': {}}

        image1.locations.extend([location3])
        image1.locations.remove(location2)

        self.assertEqual([location3], image_stub1.locations)
        self.assertEqual([location3], image1.locations)
        self.assertRaises(ValueError,
                          image1.locations.remove, location_bad)

        image1.delete()
        image2.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())
        self.assertNotIn(UUID3, self.store_api.data.keys())

    def test_image_change_delete_location(self):
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)

        del image1.locations[0]

        self.assertEqual([], image_stub1.locations)
        self.assertEqual(0, len(image1.locations))

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())

        image1.delete()

    def test_image_change_insert_invalid_location_uri(self):
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)

        location_bad = {'url': 'unknown://location', 'metadata': {}}
        self.assertRaises(exception.BadStoreUri,
                          image1.locations.insert, 0, location_bad)

        image1.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())

    def test_image_change_insert_invalid_location_metadata(self):
        UUID3 = 'a8a61ec4-d7a3-11e2-8c28-000c29c27581'
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)
        (image2, image_stub2) = self._add_image(context, UUID3, 'YYYY', 4)

        location_bad = {'url': UUID3, 'metadata': b"a invalid metadata"}

        self.assertRaises(glance_store.BackendException,
                          image1.locations.insert, 0, location_bad)

        image1.delete()
        image2.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())
        self.assertNotIn(UUID3, self.store_api.data.keys())

    def test_image_change_insert_location(self):
        UUID3 = 'a8a61ec4-d7a3-11e2-8c28-000c29c27581'
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)
        (image2, image_stub2) = self._add_image(context, UUID3, 'YYYY', 4)

        location2 = {'url': UUID2, 'metadata': {}, 'status': 'active'}
        location3 = {'url': UUID3, 'metadata': {}, 'status': 'active'}

        image1.locations.insert(0, location3)

        self.assertEqual([location3, location2], image_stub1.locations)
        self.assertEqual([location3, location2], image1.locations)

        image1.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())
        self.assertNotIn(UUID3, self.store_api.data.keys())

        image2.delete()

    def test_image_change_delete_locations(self):
        UUID3 = 'a8a61ec4-d7a3-11e2-8c28-000c29c27581'
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)
        (image2, image_stub2) = self._add_image(context, UUID3, 'YYYY', 4)

        location2 = {'url': UUID2, 'metadata': {}}
        location3 = {'url': UUID3, 'metadata': {}}

        image1.locations.insert(0, location3)
        del image1.locations[0:100]

        self.assertEqual([], image_stub1.locations)
        self.assertEqual(0, len(image1.locations))
        self.assertRaises(exception.BadStoreUri,
                          image1.locations.insert, 0, location2)
        self.assertRaises(exception.BadStoreUri,
                          image2.locations.insert, 0, location3)

        image1.delete()
        image2.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())
        self.assertNotIn(UUID3, self.store_api.data.keys())

    def test_image_change_adding_invalid_location_uri(self):
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        image_stub1 = ImageStub('fake_image_id', status='queued', locations=[])
        image1 = glance.location.ImageProxy(image_stub1, context,
                                            self.store_api, self.store_utils)

        location_bad = {'url': 'unknown://location', 'metadata': {}}

        self.assertRaises(exception.BadStoreUri,
                          image1.locations.__iadd__, [location_bad])
        self.assertEqual([], image_stub1.locations)
        self.assertEqual([], image1.locations)

        image1.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())

    def test_image_change_adding_invalid_location_metadata(self):
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)

        image_stub2 = ImageStub('fake_image_id', status='queued', locations=[])
        image2 = glance.location.ImageProxy(image_stub2, context,
                                            self.store_api, self.store_utils)

        location_bad = {'url': UUID2, 'metadata': b"a invalid metadata"}

        self.assertRaises(glance_store.BackendException,
                          image2.locations.__iadd__, [location_bad])
        self.assertEqual([], image_stub2.locations)
        self.assertEqual([], image2.locations)

        image1.delete()
        image2.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())

    def test_image_change_adding_locations(self):
        UUID3 = 'a8a61ec4-d7a3-11e2-8c28-000c29c27581'
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)
        (image2, image_stub2) = self._add_image(context, UUID3, 'YYYY', 4)

        image_stub3 = ImageStub('fake_image_id', status='queued', locations=[])
        image3 = glance.location.ImageProxy(image_stub3, context,
                                            self.store_api, self.store_utils)

        location2 = {'url': UUID2, 'metadata': {}}
        location3 = {'url': UUID3, 'metadata': {}}
        with mock.patch("glance.location.store") as mock_store:
            mock_store.get_size_from_uri_and_backend.return_value = 4
            image3.locations += [location2, location3]

        self.assertEqual([location2, location3], image_stub3.locations)
        self.assertEqual([location2, location3], image3.locations)

        image3.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())
        self.assertNotIn(UUID3, self.store_api.data.keys())

        image1.delete()
        image2.delete()

    def test_image_get_location_index(self):
        UUID3 = 'a8a61ec4-d7a3-11e2-8c28-000c29c27581'
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)
        (image2, image_stub2) = self._add_image(context, UUID3, 'YYYY', 4)
        image_stub3 = ImageStub('fake_image_id', status='queued', locations=[])

        image3 = glance.location.ImageProxy(image_stub3, context,
                                            self.store_api, self.store_utils)

        location2 = {'url': UUID2, 'metadata': {}}
        location3 = {'url': UUID3, 'metadata': {}}

        with mock.patch("glance.location.store") as mock_store:
            mock_store.get_size_from_uri_and_backend.return_value = 4
            image3.locations += [location2, location3]

        self.assertEqual(1, image_stub3.locations.index(location3))

        image3.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())
        self.assertNotIn(UUID3, self.store_api.data.keys())

        image1.delete()
        image2.delete()

    def test_image_get_location_by_index(self):
        UUID3 = 'a8a61ec4-d7a3-11e2-8c28-000c29c27581'
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)
        (image2, image_stub2) = self._add_image(context, UUID3, 'YYYY', 4)
        image_stub3 = ImageStub('fake_image_id', status='queued', locations=[])
        image3 = glance.location.ImageProxy(image_stub3, context,
                                            self.store_api, self.store_utils)

        location2 = {'url': UUID2, 'metadata': {}}
        location3 = {'url': UUID3, 'metadata': {}}

        with mock.patch("glance.location.store") as mock_store:
            mock_store.get_size_from_uri_and_backend.return_value = 4
            image3.locations += [location2, location3]

        self.assertEqual(1, image_stub3.locations.index(location3))
        self.assertEqual(location2, image_stub3.locations[0])

        image3.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())
        self.assertNotIn(UUID3, self.store_api.data.keys())

        image1.delete()
        image2.delete()

    def test_image_checking_location_exists(self):
        UUID3 = 'a8a61ec4-d7a3-11e2-8c28-000c29c27581'
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)
        (image2, image_stub2) = self._add_image(context, UUID3, 'YYYY', 4)

        image_stub3 = ImageStub('fake_image_id', status='queued', locations=[])
        image3 = glance.location.ImageProxy(image_stub3, context,
                                            self.store_api, self.store_utils)

        location2 = {'url': UUID2, 'metadata': {}}
        location3 = {'url': UUID3, 'metadata': {}}
        location_bad = {'url': 'unknown://location', 'metadata': {}}

        with mock.patch("glance.location.store") as mock_store:
            mock_store.get_size_from_uri_and_backend.return_value = 4
            image3.locations += [location2, location3]

        self.assertIn(location3, image_stub3.locations)
        self.assertNotIn(location_bad, image_stub3.locations)

        image3.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())
        self.assertNotIn(UUID3, self.store_api.data.keys())

        image1.delete()
        image2.delete()

    def test_image_reverse_locations_order(self):
        UUID3 = 'a8a61ec4-d7a3-11e2-8c28-000c29c27581'
        self.assertEqual(2, len(self.store_api.data.keys()))

        context = glance.context.RequestContext(user=USER1)
        (image1, image_stub1) = self._add_image(context, UUID2, 'XXXX', 4)
        (image2, image_stub2) = self._add_image(context, UUID3, 'YYYY', 4)

        location2 = {'url': UUID2, 'metadata': {}}
        location3 = {'url': UUID3, 'metadata': {}}

        image_stub3 = ImageStub('fake_image_id', status='queued', locations=[])
        image3 = glance.location.ImageProxy(image_stub3, context,
                                            self.store_api, self.store_utils)
        with mock.patch("glance.location.store") as mock_store:
            mock_store.get_size_from_uri_and_backend.return_value = 4
            image3.locations += [location2, location3]

        image_stub3.locations.reverse()

        self.assertEqual([location3, location2], image_stub3.locations)
        self.assertEqual([location3, location2], image3.locations)

        image3.delete()

        self.assertEqual(2, len(self.store_api.data.keys()))
        self.assertNotIn(UUID2, self.store_api.data.keys())
        self.assertNotIn(UUID3, self.store_api.data.keys())

        image1.delete()
        image2.delete()


class TestStoreImageRepo(utils.BaseTestCase):
    def setUp(self):
        super(TestStoreImageRepo, self).setUp()
        self.store_api = unit_test_utils.FakeStoreAPI()
        store_utils = unit_test_utils.FakeStoreUtils(self.store_api)
        self.image_stub = ImageStub(UUID1)
        self.image = glance.location.ImageProxy(self.image_stub, {},
                                                self.store_api, store_utils)
        self.image_repo_stub = ImageRepoStub()
        self.image_repo = glance.location.ImageRepoProxy(self.image_repo_stub,
                                                         {}, self.store_api,
                                                         store_utils)
        patcher = mock.patch("glance.location._get_member_repo_for_store",
                             self.get_fake_member_repo)
        patcher.start()
        self.addCleanup(patcher.stop)
        self.fake_member_repo = FakeMemberRepo(self.image, [TENANT1, TENANT2])
        self.image_member_repo = glance.location.ImageMemberRepoProxy(
            self.fake_member_repo,
            self.image,
            {}, self.store_api)

    def get_fake_member_repo(self, image, context, db_api, store_api):
        return FakeMemberRepo(self.image, [TENANT1, TENANT2])

    def test_add_updates_acls(self):
        self.image_stub.locations = [{'url': 'foo', 'metadata': {},
                                      'status': 'active'},
                                     {'url': 'bar', 'metadata': {},
                                      'status': 'active'}]
        self.image_stub.visibility = 'public'
        self.image_repo.add(self.image)
        self.assertTrue(self.store_api.acls['foo']['public'])
        self.assertEqual([], self.store_api.acls['foo']['read'])
        self.assertEqual([], self.store_api.acls['foo']['write'])
        self.assertTrue(self.store_api.acls['bar']['public'])
        self.assertEqual([], self.store_api.acls['bar']['read'])
        self.assertEqual([], self.store_api.acls['bar']['write'])

    def test_add_ignores_acls_if_no_locations(self):
        self.image_stub.locations = []
        self.image_stub.visibility = 'public'
        self.image_repo.add(self.image)
        self.assertEqual(0, len(self.store_api.acls))

    def test_save_updates_acls(self):
        self.image_stub.locations = [{'url': 'foo', 'metadata': {},
                                      'status': 'active'}]
        self.image_repo.save(self.image)
        self.assertIn('foo', self.store_api.acls)

    def test_add_fetches_members_if_private(self):
        self.image_stub.locations = [{'url': 'glue', 'metadata': {},
                                      'status': 'active'}]
        self.image_stub.visibility = 'private'
        self.image_repo.add(self.image)
        self.assertIn('glue', self.store_api.acls)
        acls = self.store_api.acls['glue']
        self.assertFalse(acls['public'])
        self.assertEqual([], acls['write'])
        self.assertEqual([TENANT1, TENANT2], acls['read'])

    def test_save_fetches_members_if_private(self):
        self.image_stub.locations = [{'url': 'glue', 'metadata': {},
                                      'status': 'active'}]
        self.image_stub.visibility = 'private'
        self.image_repo.save(self.image)
        self.assertIn('glue', self.store_api.acls)
        acls = self.store_api.acls['glue']
        self.assertFalse(acls['public'])
        self.assertEqual([], acls['write'])
        self.assertEqual([TENANT1, TENANT2], acls['read'])

    def test_member_addition_updates_acls(self):
        self.image_stub.locations = [{'url': 'glug', 'metadata': {},
                                      'status': 'active'}]
        self.image_stub.visibility = 'private'
        membership = glance.domain.ImageMembership(
            UUID1, TENANT3, None, None, status='accepted')
        self.image_member_repo.add(membership)
        self.assertIn('glug', self.store_api.acls)
        acls = self.store_api.acls['glug']
        self.assertFalse(acls['public'])
        self.assertEqual([], acls['write'])
        self.assertEqual([TENANT1, TENANT2, TENANT3], acls['read'])

    def test_member_removal_updates_acls(self):
        self.image_stub.locations = [{'url': 'glug', 'metadata': {},
                                      'status': 'active'}]
        self.image_stub.visibility = 'private'
        membership = glance.domain.ImageMembership(
            UUID1, TENANT1, None, None, status='accepted')
        self.image_member_repo.remove(membership)
        self.assertIn('glug', self.store_api.acls)
        acls = self.store_api.acls['glug']
        self.assertFalse(acls['public'])
        self.assertEqual([], acls['write'])
        self.assertEqual([TENANT2], acls['read'])


class TestImageFactory(unit_test_base.StoreClearingUnitTest):

    def setUp(self):
        super(TestImageFactory, self).setUp()
        store_api = unit_test_utils.FakeStoreAPI()
        store_utils = unit_test_utils.FakeStoreUtils(store_api)
        self.image_factory = glance.location.ImageFactoryProxy(
            ImageFactoryStub(),
            glance.context.RequestContext(user=USER1),
            store_api,
            store_utils)

    def test_new_image(self):
        image = self.image_factory.new_image()
        self.assertIsNone(image.image_id)
        self.assertIsNone(image.status)
        self.assertEqual('private', image.visibility)
        self.assertEqual([], image.locations)

    def test_new_image_with_location(self):
        locations = [{'url': '%s/%s' % (BASE_URI, UUID1),
                      'metadata': {}}]
        image = self.image_factory.new_image(locations=locations)
        self.assertEqual(locations, image.locations)
        location_bad = {'url': 'unknown://location', 'metadata': {}}
        self.assertRaises(exception.BadStoreUri,
                          self.image_factory.new_image,
                          locations=[location_bad])


class TestStoreMetaDataChecker(utils.BaseTestCase):

    def test_empty(self):
        glance_store.check_location_metadata({})

    def test_unicode(self):
        m = {'key': 'somevalue'}
        glance_store.check_location_metadata(m)

    def test_unicode_list(self):
        m = {'key': ['somevalue', '2']}
        glance_store.check_location_metadata(m)

    def test_unicode_dict(self):
        inner = {'key1': 'somevalue', 'key2': 'somevalue'}
        m = {'topkey': inner}
        glance_store.check_location_metadata(m)

    def test_unicode_dict_list(self):
        inner = {'key1': 'somevalue', 'key2': 'somevalue'}
        m = {'topkey': inner, 'list': ['somevalue', '2'], 'u': '2'}
        glance_store.check_location_metadata(m)

    def test_nested_dict(self):
        inner = {'key1': 'somevalue', 'key2': 'somevalue'}
        inner = {'newkey': inner}
        inner = {'anotherkey': inner}
        m = {'topkey': inner}
        glance_store.check_location_metadata(m)

    def test_simple_bad(self):
        m = {'key1': object()}
        self.assertRaises(glance_store.BackendException,
                          glance_store.check_location_metadata,
                          m)

    def test_list_bad(self):
        m = {'key1': ['somevalue', object()]}
        self.assertRaises(glance_store.BackendException,
                          glance_store.check_location_metadata,
                          m)

    def test_nested_dict_bad(self):
        inner = {'key1': 'somevalue', 'key2': object()}
        inner = {'newkey': inner}
        inner = {'anotherkey': inner}
        m = {'topkey': inner}

        self.assertRaises(glance_store.BackendException,
                          glance_store.check_location_metadata,
                          m)



# Copyright 2016 Rackspace
# Copyright 2013 Intel Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from alembic import op
from sqlalchemy.schema import (
    Column, PrimaryKeyConstraint, ForeignKeyConstraint)

from glance.db.sqlalchemy.schema import (
    Boolean, DateTime, String, Text)  # noqa
from glance.db.sqlalchemy.models import JSONEncodedDict


def _add_tasks_table():
    op.create_table('tasks',
                    Column('id', String(length=36), nullable=False),
                    Column('type', String(length=30), nullable=False),
                    Column('status', String(length=30), nullable=False),
                    Column('owner', String(length=255), nullable=False),
                    Column('expires_at', DateTime(), nullable=True),
                    Column('created_at', DateTime(), nullable=False),
                    Column('updated_at', DateTime(), nullable=True),
                    Column('deleted_at', DateTime(), nullable=True),
                    Column('deleted', Boolean(), nullable=False),
                    PrimaryKeyConstraint('id'),
                    mysql_engine='InnoDB',
                    mysql_charset='utf8',
                    extend_existing=True)

    op.create_index('ix_tasks_deleted', 'tasks', ['deleted'], unique=False)
    op.create_index('ix_tasks_owner', 'tasks', ['owner'], unique=False)
    op.create_index('ix_tasks_status', 'tasks', ['status'], unique=False)
    op.create_index('ix_tasks_type', 'tasks', ['type'], unique=False)
    op.create_index('ix_tasks_updated_at',
                    'tasks',
                    ['updated_at'],
                    unique=False)


def _add_task_info_table():
    op.create_table('task_info',
                    Column('task_id', String(length=36), nullable=False),
                    Column('input', JSONEncodedDict(), nullable=True),
                    Column('result', JSONEncodedDict(), nullable=True),
                    Column('message', Text(), nullable=True),
                    ForeignKeyConstraint(['task_id'], ['tasks.id'], ),
                    PrimaryKeyConstraint('task_id'),
                    mysql_engine='InnoDB',
                    mysql_charset='utf8',
                    extend_existing=True)


def upgrade():
    _add_tasks_table()
    _add_task_info_table()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""drop glare artifacts tables

Revision ID: pike_contract01
Revises: ocata_contract01
Create Date: 2017-02-09 20:32:51.222867

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = 'pike_contract01'
down_revision = 'ocata_contract01'
branch_labels = None
depends_on = 'pike_expand01'


def upgrade():
    # create list of artifact tables in reverse order of their creation
    table_names = []
    table_names.append('artifact_blob_locations')
    table_names.append('artifact_properties')
    table_names.append('artifact_blobs')
    table_names.append('artifact_dependencies')
    table_names.append('artifact_tags')
    table_names.append('artifacts')

    for table_name in table_names:
        op.drop_table(table_name=table_name)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from oslo_db.sqlalchemy import test_fixtures
from oslo_db.sqlalchemy import utils as db_utils

from glance.tests.functional.db import test_migrations
import glance.tests.utils as test_utils


class TestMitaka02Mixin(test_migrations.AlembicMigrationsMixin):

    def _pre_upgrade_mitaka02(self, engine):
        metadef_resource_types = db_utils.get_table(
            engine, 'metadef_resource_types')
        now = datetime.datetime.now()
        db_rec1 = dict(id='9580',
                       name='OS::Nova::Instance',
                       protected=False,
                       created_at=now,
                       updated_at=now,)
        db_rec2 = dict(id='9581',
                       name='OS::Nova::Blah',
                       protected=False,
                       created_at=now,
                       updated_at=now,)
        db_values = (db_rec1, db_rec2)
        with engine.connect() as conn, conn.begin():
            conn.execute(metadef_resource_types.insert().values(db_values))

    def _check_mitaka02(self, engine, data):
        metadef_resource_types = db_utils.get_table(
            engine, 'metadef_resource_types')
        with engine.connect() as conn:
            result = conn.execute(
                metadef_resource_types.select()
                .where(metadef_resource_types.c.name == 'OS::Nova::Instance')
            ).fetchall()
            self.assertEqual(0, len(result))

            result = conn.execute(
                metadef_resource_types.select()
                .where(metadef_resource_types.c.name == 'OS::Nova::Server')
            ).fetchall()
            self.assertEqual(1, len(result))


class TestMitaka02MySQL(
    TestMitaka02Mixin,
    test_fixtures.OpportunisticDBTestMixin,
    test_utils.BaseTestCase,
):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture


class TestMitaka02PostgresSQL(
    TestMitaka02Mixin,
    test_fixtures.OpportunisticDBTestMixin,
    test_utils.BaseTestCase,
):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture


class TestMitaka02Sqlite(
    TestMitaka02Mixin,
    test_fixtures.OpportunisticDBTestMixin,
    test_utils.BaseTestCase,
):
    pass



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import http.client as http

from oslo_serialization import jsonutils

from glance.tests.functional.v2 import metadef_base


class TestNamespaceProperties(metadef_base.MetadefFunctionalTestBase):

    def setUp(self):
        super(TestNamespaceProperties, self).setUp()
        self.start_server(enable_cache=False)

    def test_properties_lifecycle(self):
        # Namespace should not exist
        path = '/v2/metadefs/namespaces/MyNamespace'
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.NOT_FOUND, response.status_code)

        # Create a namespace
        path = '/v2/metadefs/namespaces'
        headers = self._headers({'content-type': 'application/json'})
        namespace_name = 'MyNamespace'
        resource_type_name = 'MyResourceType'
        resource_type_prefix = 'MyPrefix'
        data = {
            "namespace": namespace_name,
            "display_name": "My User Friendly Namespace",
            "description": "My description",
            "visibility": "public",
            "protected": False,
            "owner": "The Test Owner",
            "resource_type_associations": [
                {
                    "name": resource_type_name,
                    "prefix": resource_type_prefix
                }
            ]
        }
        response = self.api_post(path, headers=headers, json=data)
        self.assertEqual(http.CREATED, response.status_code)

        # Property1 should not exist
        path = '/v2/metadefs/namespaces/MyNamespace/properties/property1'
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.NOT_FOUND, response.status_code)

        # Create a property
        path = '/v2/metadefs/namespaces/MyNamespace/properties'
        headers = self._headers({'content-type': 'application/json'})
        property_name = "property1"
        data = {
            "name": property_name,
            "type": "integer",
            "title": "property1",
            "description": "property1 description",
            "default": 100,
            "minimum": 100,
            "maximum": 30000369,
            "readonly": False,
        }
        response = self.api_post(path, headers=headers, json=data)
        self.assertEqual(http.CREATED, response.status_code)

        # Attempt to insert a duplicate
        response = self.api_post(path, headers=headers, json=data)
        self.assertEqual(http.CONFLICT, response.status_code)

        # Get the property created above
        path = '/v2/metadefs/namespaces/%s/properties/%s' % (
            namespace_name, property_name)
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        property_object = jsonutils.loads(response.text)
        self.assertEqual("integer", property_object['type'])
        self.assertEqual("property1", property_object['title'])
        self.assertEqual("property1 description", property_object[
            'description'])
        self.assertEqual('100', property_object['default'])
        self.assertEqual(100, property_object['minimum'])
        self.assertEqual(30000369, property_object['maximum'])

        # Get the property with specific resource type association
        path = '/v2/metadefs/namespaces/%s/properties/%s%s' % (
            namespace_name, property_name, '='.join(['?resource_type',
                                                    resource_type_name]))
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.NOT_FOUND, response.status_code)

        # Get the property with prefix and specific resource type association
        property_name_with_prefix = ''.join([resource_type_prefix,
                                            property_name])
        path = '/v2/metadefs/namespaces/%s/properties/%s%s' % (
            namespace_name, property_name_with_prefix, '='.join([
                '?resource_type', resource_type_name]))
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.OK, response.status_code)
        property_object = jsonutils.loads(response.text)
        self.assertEqual("integer", property_object['type'])
        self.assertEqual("property1", property_object['title'])
        self.assertEqual("property1 description", property_object[
            'description'])
        self.assertEqual('100', property_object['default'])
        self.assertEqual(100, property_object['minimum'])
        self.assertEqual(30000369, property_object['maximum'])
        self.assertFalse(property_object['readonly'])

        # Returned property should match the created property
        property_object = jsonutils.loads(response.text)
        checked_keys = set([
            'name',
            'type',
            'title',
            'description',
            'default',
            'minimum',
            'maximum',
            'readonly',
        ])
        self.assertEqual(set(property_object.keys()), checked_keys)
        expected_metadata_property = {
            "type": "integer",
            "title": "property1",
            "description": "property1 description",
            "default": '100',
            "minimum": 100,
            "maximum": 30000369,
            "readonly": False,
        }

        for key, value in expected_metadata_property.items():
            self.assertEqual(property_object[key], value, key)

        # The property should be mutable
        path = '/v2/metadefs/namespaces/%s/properties/%s' % (
            namespace_name, property_name)
        media_type = 'application/json'
        headers = self._headers({'content-type': media_type})
        property_name = "property1-UPDATED"
        data = {
            "name": property_name,
            "type": "string",
            "title": "string property",
            "description": "desc-UPDATED",
            "operators": ["<or>"],
            "default": "value-UPDATED",
            "minLength": 5,
            "maxLength": 10,
            "readonly": True,
        }
        response = self.api_put(path, headers=headers, json=data)
        self.assertEqual(http.OK, response.status_code, response.text)

        # Returned property should reflect the changes
        property_object = jsonutils.loads(response.text)
        self.assertEqual('string', property_object['type'])
        self.assertEqual('desc-UPDATED', property_object['description'])
        self.assertEqual('value-UPDATED', property_object['default'])
        self.assertEqual(["<or>"], property_object['operators'])
        self.assertEqual(5, property_object['minLength'])
        self.assertEqual(10, property_object['maxLength'])
        self.assertTrue(property_object['readonly'])

        # Updates should persist across requests
        path = '/v2/metadefs/namespaces/%s/properties/%s' % (
            namespace_name, property_name)
        response = self.api_get(path, headers=self._headers())
        property_object = jsonutils.loads(response.text)
        self.assertEqual('string', property_object['type'])
        self.assertEqual('desc-UPDATED', property_object['description'])
        self.assertEqual('value-UPDATED', property_object['default'])
        self.assertEqual(["<or>"], property_object['operators'])
        self.assertEqual(5, property_object['minLength'])
        self.assertEqual(10, property_object['maxLength'])

        # Deletion of property property1
        path = '/v2/metadefs/namespaces/%s/properties/%s' % (
            namespace_name, property_name)
        response = self.api_delete(path, headers=self._headers())
        self.assertEqual(http.NO_CONTENT, response.status_code)

        # property1 should not exist
        path = '/v2/metadefs/namespaces/%s/properties/%s' % (
            namespace_name, property_name)
        response = self.api_get(path, headers=self._headers())
        self.assertEqual(http.NOT_FOUND, response.status_code)

    def _create_properties(self, namespaces):
        properties = []
        for namespace in namespaces:
            headers = self._headers({'X-Tenant-Id': namespace['owner']})
            data = {
                "name": "property_of_%s" % (namespace['namespace']),
                "type": "integer",
                "title": "property",
                "description": "property description",
            }
            path = ('/v2/metadefs/namespaces/%s/'
                    'properties') % namespace['namespace']
            response = self.api_post(path, headers=headers, json=data)
            self.assertEqual(http.CREATED, response.status_code)
            prop_metadata = jsonutils.loads(response.text)
            metadef_property = dict()
            metadef_property[namespace['namespace']] = prop_metadata['name']
            properties.append(metadef_property)

        return properties

    def _update_property(self, path, headers, data):
        # The property should be mutable
        response = self.api_put(path, headers=headers, json=data)
        self.assertEqual(http.OK, response.status_code, response.text)

        # Returned property should reflect the changes
        property_object = jsonutils.loads(response.text)
        self.assertEqual('string', property_object['type'])
        self.assertEqual(data['description'], property_object['description'])

        # Updates should persist across requests
        response = self.api_get(path, headers=self._headers())
        property_object = jsonutils.loads(response.text)
        self.assertEqual('string', property_object['type'])
        self.assertEqual(data['description'], property_object['description'])

    def test_role_base_metadata_properties_lifecycle(self):
        # Create public and private namespaces for tenant1 and tenant2
        path = '/v2/metadefs/namespaces'
        headers = self._headers({'content-type': 'application/json'})
        tenant1_namespaces = []
        tenant2_namespaces = []
        for tenant in [self.tenant1, self.tenant2]:
            headers['X-Tenant-Id'] = tenant
            for visibility in ['public', 'private']:
                namespace_data = {
                    "namespace": "%s_%s_namespace" % (tenant, visibility),
                    "display_name": "My User Friendly Namespace",
                    "description": "My description",
                    "visibility": visibility,
                    "owner": tenant
                }
                namespace = self.create_namespace(path, headers,
                                                  namespace_data)
                namespace = jsonutils.loads(namespace.text)
                self.assertNamespacesEqual(namespace, namespace_data)
                if tenant == self.tenant1:
                    tenant1_namespaces.append(namespace)
                else:
                    tenant2_namespaces.append(namespace)

        # Create a metadef property for each namespace created above
        tenant1_properties = self._create_properties(tenant1_namespaces)
        tenant2_properties = self._create_properties(tenant2_namespaces)

        def _check_properties_access(properties, tenant):
            headers = self._headers({'content-type': 'application/json',
                                     'X-Tenant-Id': tenant,
                                     'X-Roles': 'reader,member'})
            for prop in properties:
                for namespace, property_name in prop.items():
                    path = '/v2/metadefs/namespaces/%s/properties/%s' % (
                        namespace, property_name)
                    response = self.api_get(path, headers=headers)
                    if namespace.split('_')[1] == 'public':
                        expected = http.OK
                    else:
                        expected = http.NOT_FOUND

                    # Make sure we can see our and public properties, but not
                    # the other tenant's
                    self.assertEqual(expected, response.status_code)

                    # Make sure the same holds for listing
                    path = '/v2/metadefs/namespaces/%s/properties' % namespace
                    response = self.api_get(path, headers=headers)
                    self.assertEqual(expected, response.status_code)
                    if expected == http.OK:
                        resp_props = jsonutils.loads(
                            response.text)['properties'].values()
                        self.assertEqual(
                            sorted(prop.values()),
                            sorted([x['name']
                                    for x in resp_props]))

        # Check Tenant 1 can access properties of all public namespace
        # and cannot access properties of private namespace of Tenant 2
        _check_properties_access(tenant2_properties, self.tenant1)

        # Check Tenant 2 can access properties of public namespace and
        # cannot access properties of private namespace of Tenant 1
        _check_properties_access(tenant1_properties, self.tenant2)

        # Update properties with admin and non admin role
        total_properties = tenant1_properties + tenant2_properties
        for prop in total_properties:
            for namespace, property_name in prop.items():
                data = {
                    "name": property_name,
                    "type": "string",
                    "title": "string property",
                    "description": "desc-UPDATED",
                }
                path = '/v2/metadefs/namespaces/%s/properties/%s' % (
                    namespace, property_name)

                # Update property should fail with non admin role
                headers['X-Roles'] = "reader,member"
                response = self.api_put(path, headers=headers, json=data)
                self.assertEqual(http.FORBIDDEN, response.status_code)

                # Should work with admin role
                headers = self._headers({
                    'X-Tenant-Id': namespace.split('_')[0]})
                self._update_property(path, headers, data)

        # Delete property should not be allowed to non admin role
        for prop in total_properties:
            for namespace, property_name in prop.items():
                path = '/v2/metadefs/namespaces/%s/properties/%s' % (
                    namespace, property_name)
                response = self.api_delete(
                    path, headers=self._headers({
                        'X-Roles': 'reader,member',
                        'X-Tenant-Id': namespace.split('_')[0]
                    }))
                self.assertEqual(http.FORBIDDEN, response.status_code)

        # Delete all metadef properties
        headers = self._headers()
        for prop in total_properties:
            for namespace, property_name in prop.items():
                path = '/v2/metadefs/namespaces/%s/properties/%s' % (
                    namespace, property_name)
                response = self.api_delete(path, headers=headers)
                self.assertEqual(http.NO_CONTENT, response.status_code)

                # Deleted property should not exist
                response = self.api_get(path, headers=headers)
                self.assertEqual(http.NOT_FOUND, response.status_code)
